# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import pathlib
from unittest import mock

import fixtures
import pytest
from testscenarios import multiply_scenarios
from testtools.matchers import Contains, Equals, FileExists, HasLength, LessThan, Not

from snapcraft import repo
from snapcraft.internal import errors
from snapcraft.plugins.v1 import _ros, colcon
from tests import unit

from . import PluginsV1BaseTestCase


class ColconPluginTestBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class props:
            colcon_rosdistro = "crystal"
            colcon_packages = ["my_package"]
            colcon_packages_ignore = []
            colcon_source_space = "src"
            source_subdir = None
            colcon_cmake_args = []
            colcon_catkin_cmake_args = []
            colcon_ament_cmake_args = []
            build_attributes = []
            disable_parallel = False

        self.properties = props()
        self.ros_version = "2"
        self.ubuntu_distro = "bionic"

        self.ubuntu_mock = self.useFixture(
            fixtures.MockPatch("snapcraft.repo.Ubuntu")
        ).mock

        self.dependencies_mock = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.plugins.v1.colcon._find_system_dependencies", return_value={}
            )
        ).mock

        self.rosdep_mock = self.useFixture(
            fixtures.MockPatch("snapcraft.plugins.v1._ros.rosdep.Rosdep")
        ).mock

        self.pip_mock = self.useFixture(
            fixtures.MockPatch("snapcraft.plugins.v1._python.Pip")
        ).mock
        self.pip_mock.return_value.list.return_value = {}

    def assert_rosdep_setup(
        self,
        rosdistro,
        ros_version,
        package_path,
        rosdep_path,
        ubuntu_distro,
        base,
        target_arch,
    ):
        self.rosdep_mock.assert_has_calls(
            [
                mock.call(
                    ros_distro=rosdistro,
                    ros_version=ros_version,
                    ros_package_path=package_path,
                    rosdep_path=rosdep_path,
                    ubuntu_distro=ubuntu_distro,
                    base=base,
                    target_arch=target_arch,
                ),
                mock.call().setup(),
            ]
        )

    def assert_pip_setup(self, python_major_version, part_dir, install_dir, stage_dir):
        self.pip_mock.assert_has_calls(
            [
                mock.call(
                    python_major_version=python_major_version,
                    part_dir=part_dir,
                    install_dir=install_dir,
                    stage_dir=stage_dir,
                ),
                mock.call().setup(),
            ]
        )


class ColconPluginTest(ColconPluginTestBase):
    def test_schema(self):
        schema = colcon.ColconPlugin.schema()

        properties = schema["properties"]
        expected = (
            "colcon-rosdistro",
            "colcon-packages",
            "colcon-source-space",
            "colcon-cmake-args",
            "colcon-catkin-cmake-args",
            "colcon-ament-cmake-args",
            "colcon-packages-ignore",
        )
        self.assertThat(properties, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(properties, Contains(prop))

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            colcon.ColconPlugin,
            "test-part",
            self.properties,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))

    def test_unsupported_base_and_rosdistro(self):
        self.project._snap_meta.base = "core"

        raised = self.assertRaises(
            colcon.ColconPluginBaseError,
            colcon.ColconPlugin,
            "test-part",
            self.properties,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("core"))
        self.assertThat(raised.rosdistro, Equals("crystal"))

    def test_schema_colcon_rosdistro(self):
        schema = colcon.ColconPlugin.schema()

        # Check colcon-rosdistro property
        colcon_packages = schema["properties"]["colcon-rosdistro"]
        expected = ("type", "default", "enum")
        self.assertThat(colcon_packages, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(colcon_packages, Contains(prop))
        self.assertThat(colcon_packages["type"], Equals("string"))
        self.assertThat(colcon_packages["default"], Equals("crystal"))

        enum = colcon_packages["enum"]
        self.assertThat(enum, HasLength(4))
        self.assertThat(enum, Contains("bouncy"))
        self.assertThat(enum, Contains("crystal"))
        self.assertThat(enum, Contains("dashing"))
        self.assertThat(enum, Contains("eloquent"))

    def test_schema_colcon_packages(self):
        schema = colcon.ColconPlugin.schema()

        # Check colcon-packages property
        colcon_packages = schema["properties"]["colcon-packages"]
        expected = ("type", "minitems", "uniqueItems", "items")
        self.assertThat(colcon_packages, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(colcon_packages, Contains(prop))
        self.assertThat(colcon_packages["type"], Equals("array"))
        self.assertThat(colcon_packages["minitems"], Equals(1))
        self.assertTrue(colcon_packages["uniqueItems"])
        self.assertThat(colcon_packages["items"], Contains("type"))
        self.assertThat(colcon_packages["items"]["type"], Equals("string"))

    def test_schema_source_space(self):
        schema = colcon.ColconPlugin.schema()

        # Check colcon-source-space property
        source_space = schema["properties"]["colcon-source-space"]
        expected = ("type", "default")
        self.assertThat(source_space, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(source_space, Contains(prop))
        self.assertThat(source_space["type"], Equals("string"))
        self.assertThat(source_space["default"], Equals("src"))

    def test_schema_colcon_cmake_args(self):
        schema = colcon.ColconPlugin.schema()

        # Check colcon-cmake-args property
        colcon_cmake_args = schema["properties"]["colcon-cmake-args"]
        expected = ("type", "default", "minitems", "items")
        self.assertThat(colcon_cmake_args, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(colcon_cmake_args, Contains(prop))
        self.assertThat(colcon_cmake_args["type"], Equals("array"))
        self.assertThat(colcon_cmake_args["default"], Equals([]))
        self.assertThat(colcon_cmake_args["minitems"], Equals(1))
        self.assertThat(colcon_cmake_args["items"], Contains("type"))
        self.assertThat(colcon_cmake_args["items"]["type"], Equals("string"))

    def test_schema_colcon_catkin_cmake_args(self):
        schema = colcon.ColconPlugin.schema()

        # Check colcon-catkin-cmake-args property
        colcon_cmake_args = schema["properties"]["colcon-catkin-cmake-args"]
        expected = ("type", "default", "minitems", "items")
        self.assertThat(colcon_cmake_args, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(colcon_cmake_args, Contains(prop))
        self.assertThat(colcon_cmake_args["type"], Equals("array"))
        self.assertThat(colcon_cmake_args["default"], Equals([]))
        self.assertThat(colcon_cmake_args["minitems"], Equals(1))
        self.assertThat(colcon_cmake_args["items"], Contains("type"))
        self.assertThat(colcon_cmake_args["items"]["type"], Equals("string"))

    def test_schema_colcon_ament_cmake_args(self):
        schema = colcon.ColconPlugin.schema()

        # Check colcon-ament-cmake-args property
        colcon_cmake_args = schema["properties"]["colcon-ament-cmake-args"]
        expected = ("type", "default", "minitems", "items")
        self.assertThat(colcon_cmake_args, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(colcon_cmake_args, Contains(prop))
        self.assertThat(colcon_cmake_args["type"], Equals("array"))
        self.assertThat(colcon_cmake_args["default"], Equals([]))
        self.assertThat(colcon_cmake_args["minitems"], Equals(1))
        self.assertThat(colcon_cmake_args["items"], Contains("type"))
        self.assertThat(colcon_cmake_args["items"]["type"], Equals("string"))

    def test_schema_colcon_packages_ignore(self):
        schema = colcon.ColconPlugin.schema()

        # check colcon-packages-ignore property
        colcon_packages_ignore = schema["properties"]["colcon-packages-ignore"]
        expected = ("type", "default", "minitems", "items", "uniqueItems")
        self.assertThat(colcon_packages_ignore, HasLength(len(expected)))
        for prop in expected:
            self.assertThat(colcon_packages_ignore, Contains(prop))
        self.assertThat(colcon_packages_ignore["type"], Equals("array"))
        self.assertThat(colcon_packages_ignore["default"], Equals([]))
        self.assertThat(colcon_packages_ignore["minitems"], Equals(1))
        self.assertTrue(colcon_packages_ignore["uniqueItems"])
        self.assertThat(colcon_packages_ignore["items"], Contains("type"))
        self.assertThat(colcon_packages_ignore["items"]["type"], Equals("string"))

    def test_eol_ros_distro_warning(self):
        self.properties.colcon_rosdistro = "crystal"

        fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(fake_logger)
        colcon.ColconPlugin("test-part", self.properties, self.project)
        self.assertThat(
            fake_logger.output,
            Contains(
                "The 'crystal' ROS distro has reached end-of-life and is no longer supported. Use at your own risk."
            ),
        )

    def test_get_pull_properties(self):
        expected_pull_properties = [
            "colcon-rosdistro",
            "colcon-packages",
            "colcon-source-space",
        ]
        actual_pull_properties = colcon.ColconPlugin.get_pull_properties()

        self.assertThat(
            actual_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, actual_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = [
            "colcon-cmake-args",
            "colcon-catkin-cmake-args",
            "colcon-ament-cmake-args",
            "colcon-packages-ignore",
        ]
        actual_build_properties = colcon.ColconPlugin.get_build_properties()

        self.assertThat(
            actual_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, actual_build_properties)

    def test_pull_invalid_dependency(self):
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"apt": {"foo"}}

        self.ubuntu_mock.fetch_stage_packages.side_effect = repo.errors.PackageNotFoundError(
            "foo"
        )

        raised = self.assertRaises(colcon.ColconAptDependencyFetchError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to fetch apt dependencies: The package 'foo' was not found."
            ),
        )

    def test_clean_pull(self):
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"apt": {"foo", "bar", "baz"}}

        plugin.pull()
        os.makedirs(plugin._rosdep_path)

        plugin.clean_pull()
        self.assertFalse(os.path.exists(plugin._rosdep_path))

    def test_valid_colcon_workspace_src(self):
        # sourcedir is expected to be the root of the Colcon workspace. Since
        # it contains a 'src' directory, this is a valid Colcon workspace.
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))
        # An exception will be raised if pull can't handle the valid workspace.
        plugin.pull()

    def test_invalid_colcon_workspace_no_src(self):
        # sourcedir is expected to be the root of the Colcon workspace. Since
        # it does not contain a `src` folder and `source-space` is 'src', this
        # should fail.
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        raised = self.assertRaises(colcon.ColconPackagePathNotFoundError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to find package path: {!r}".format(
                    os.path.join(plugin.sourcedir, "src")
                )
            ),
        )

    def test_valid_colcon_workspace_source_space(self):
        self.properties.colcon_source_space = "foo"

        # sourcedir is expected to be the root of the Colcon workspace.
        # Normally this would mean it contained a `src` directory, but it can
        # be remapped via the `colcon-source-space` key.
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, self.properties.colcon_source_space))
        # An exception will be raised if pull can't handle the source space.
        plugin.pull()

    def test_invalid_colcon_workspace_invalid_source_space(self):
        self.properties.colcon_source_space = "foo"

        # sourcedir is expected to be the root of the Colcon workspace. Since
        # it does not contain a `src` folder and source_space wasn't
        # specified, this should fail.
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        raised = self.assertRaises(colcon.ColconPackagePathNotFoundError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to find package path: {!r}".format(
                    os.path.join(plugin.sourcedir, self.properties.colcon_source_space)
                )
            ),
        )

    def test_invalid_colcon_workspace_invalid_source_space_build_all(self):
        self.properties.colcon_source_space = "foo"
        self.properties.colcon_packages = None

        # sourcedir is expected to be the root of the Colcon workspace. Since
        # it does not contain a `src` folder and source_space wasn't
        # specified, this should fail.
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        raised = self.assertRaises(colcon.ColconPackagePathNotFoundError, plugin.pull)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to find package path: {!r}".format(
                    os.path.join(plugin.sourcedir, self.properties.colcon_source_space)
                )
            ),
        )

    def test_invalid_colcon_workspace_invalid_source_no_packages(self):
        """Test that an invalid source space is fine iff no packages."""
        self.properties.colcon_source_space = "foo"
        self.properties.colcon_packages = []

        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)

        # Normally pulling should fail, but since there are no packages to
        # build, even an invalid workspace should be okay.
        plugin.pull()

    def test_invalid_colcon_workspace_source_space_same_as_source(self):
        self.properties.colcon_source_space = "."

        # sourcedir is expected to be the root of the Colcon workspace. Since
        # source_space was specified to be the same as the root, this should
        # fail.
        self.assertRaises(
            colcon.ColconWorkspaceIsRootError,
            colcon.ColconPlugin,
            "test-part",
            self.properties,
            self.project,
        )

    @mock.patch.object(colcon.ColconPlugin, "run")
    @mock.patch.object(colcon.ColconPlugin, "run_output", return_value="foo")
    @mock.patch.object(colcon.ColconPlugin, "_prepare_build")
    @mock.patch.object(colcon.ColconPlugin, "_finish_build")
    def test_build_multiple(
        self, finish_build_mock, prepare_build_mock, run_output_mock, run_mock
    ):
        self.properties.colcon_packages.append("package_2")

        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        plugin.build()

        class check_pkg_arguments:
            def __init__(self, test):
                self.test = test

            def __eq__(self, args):
                index = args.index("--packages-select")
                packages = args[index + 1 : index + 3]
                self.test.assertIn("my_package", packages)
                self.test.assertIn("package_2", packages)
                return True

        run_mock.assert_called_with(check_pkg_arguments(self))

        self.assertFalse(
            self.dependencies_mock.called,
            "Dependencies should have been discovered in the pull() step",
        )

        finish_build_mock.assert_called_once_with()

    @mock.patch.object(colcon.ColconPlugin, "run_output", return_value="bar")
    def test_run_environment(self, run_mock):
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)

        with mock.patch.object(
            plugin, "_source_setup_sh", wraps=plugin._source_setup_sh
        ) as sh_mock:
            # Joining and re-splitting to get hacked script in there as well
            environment = "\n".join(plugin.env(plugin.installdir)).split("\n")
            sh_mock.assert_called_with(plugin.installdir)

        underlay_setup = os.path.join(plugin.options.colcon_rosdistro, "setup.sh")
        overlay_setup = os.path.join("snap", "local_setup.sh")

        # Verify that the python executables and root are set before any setup.sh is
        # sourced. Also verify that the underlay setup is sourced before the overlay.
        ament_python_index = [
            i for i, line in enumerate(environment) if "AMENT_PYTHON_EXECUTABLE" in line
        ][0]
        colcon_python_index = [
            i
            for i, line in enumerate(environment)
            if "COLCON_PYTHON_EXECUTABLE" in line
        ][0]
        root_index = [
            i for i, line in enumerate(environment) if "SNAP_COLCON_ROOT" in line
        ][0]
        underlay_source_setup_index = [
            i for i, line in enumerate(environment) if underlay_setup in line
        ][0]
        overlay_source_setup_index = [
            i for i, line in enumerate(environment) if overlay_setup in line
        ][0]
        self.assertThat(ament_python_index, LessThan(underlay_source_setup_index))
        self.assertThat(colcon_python_index, LessThan(underlay_source_setup_index))
        self.assertThat(root_index, LessThan(underlay_source_setup_index))
        self.assertThat(
            underlay_source_setup_index, LessThan(overlay_source_setup_index)
        )

    def test_source_setup_sh(self):
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)

        underlay = os.path.join(
            "test-root", "opt", "ros", plugin.options.colcon_rosdistro
        )
        underlay_setup = os.path.join(underlay, "setup.sh")
        overlay = os.path.join("test-root", "opt", "ros", "snap")
        overlay_setup = os.path.join(overlay, "local_setup.sh")

        # Make sure $@ is zeroed, then setup.sh sourced, then $@ is restored
        lines_of_interest = [
            "set --",
            'if [ -f "{}" ]; then'.format(underlay_setup),
            '. "{}"'.format(underlay_setup),
            "fi",
            'if [ -f "{}" ]; then'.format(overlay_setup),
            '. "{}"'.format(overlay_setup),
            "fi",
            'eval "set -- $BACKUP_ARGS"',
        ]

        actual_lines = []

        for line in plugin._source_setup_sh("test-root").split("\n"):
            line = line.strip()
            if line in lines_of_interest:
                actual_lines.append(line)

        self.assertThat(
            actual_lines,
            Equals(lines_of_interest),
            "Expected setups to be sourced after args were zeroed, followed by the "
            "args being restored.",
        )


class PrepareBuildTest(ColconPluginTestBase):
    def setUp(self):
        super().setUp()
        self.plugin = colcon.ColconPlugin("test-part", self.properties, self.project)

    @mock.patch("snapcraft.internal.mangling.rewrite_python_shebangs")
    def test_in_snap_python_is_used(self, shebangs_mock):
        # Mangling has its own tests. Here we just need to make sure
        # _prepare_build actually uses it.
        self.plugin._prepare_build()
        shebangs_mock.assert_called_once_with(self.plugin.installdir)

    def test_underlay_cmake_paths_are_rewritten(self):
        os.makedirs(os.path.join(self.plugin._ros_underlay, "test"))

        # Place a few .cmake files with incorrect paths, and some files that
        # shouldn't be changed.
        files = [
            {
                "path": "fooConfig.cmake",
                "contents": '"/usr/lib/foo"',
                "expected": '"{}/usr/lib/foo"'.format(self.plugin.installdir),
            },
            {
                "path": "bar.cmake",
                "contents": '"/usr/lib/bar"',
                "expected": '"/usr/lib/bar"',
            },
            {
                "path": "test/bazConfig.cmake",
                "contents": '"/test/baz;/usr/lib/baz"',
                "expected": '"{0}/test/baz;{0}/usr/lib/baz"'.format(
                    self.plugin.installdir
                ),
            },
            {"path": "test/quxConfig.cmake", "contents": "qux", "expected": "qux"},
            {
                "path": "test/installedConfig.cmake",
                "contents": '"{}/foo"'.format(self.plugin.installdir),
                "expected": '"{}/foo"'.format(self.plugin.installdir),
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._prepare_build()

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    def test_cmake_paths_are_rewritten(self):
        # Place a few .cmake files with incorrect paths, and some files that
        # shouldn't be changed.
        files = [
            {
                "path": "fooConfig.cmake",
                "contents": '"/usr/lib/foo"',
                "expected": '"{}/usr/lib/foo"'.format(self.plugin.installdir),
            },
            {"path": "bar", "contents": '"/usr/lib/bar"', "expected": '"/usr/lib/bar"'},
            {
                "path": "test/bazConfig.cmake",
                "contents": '"/test/baz;/usr/lib/baz"',
                "expected": '"{0}/test/baz;{0}/usr/lib/baz"'.format(
                    self.plugin.installdir
                ),
            },
            {"path": "test/quxConfig.cmake", "contents": "qux", "expected": "qux"},
            {
                "path": "test/installedConfig.cmake",
                "contents": '"{}/foo"'.format(self.plugin.installdir),
                "expected": '"{}/foo"'.format(self.plugin.installdir),
            },
            {
                "path": "test/poco.cmake",
                "contents": 'INTERFACE_LINK_LIBRARIES "pthread;dl;rt;/usr/lib/x86_64-linux-gnu/libpcre.so;/usr/lib/x86_64-linux-gnu/libz.so"',
                "expected": 'INTERFACE_LINK_LIBRARIES "pthread;dl;rt;{0}/usr/lib/x86_64-linux-gnu/libpcre.so;{0}/usr/lib/x86_64-linux-gnu/libz.so"'.format(
                    self.plugin.installdir
                ),
            },
        ]

        for file_info in files:
            path = os.path.join(
                self.plugin.installdir, "usr", "lib", "cmake", file_info["path"]
            )
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._prepare_build()

        for file_info in files:
            path = os.path.join(
                self.plugin.installdir, "usr", "lib", "cmake", file_info["path"]
            )
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    def test_ament_prefix_is_rewritten(self):
        os.makedirs(os.path.join(self.plugin._ros_underlay, "test"))

        # Place files with incorrect prefixes, and some that shouldn't be changed.
        files = [
            {
                "path": "setup.sh",
                "contents": ": ${AMENT_CURRENT_PREFIX:=/opt/ros/crystal}",
                "expected": ': ${AMENT_CURRENT_PREFIX:="$SNAP_COLCON_ROOT/opt/ros/crystal"}',
            },
            {
                "path": "test/local_setup.sh",
                "contents": ': ${AMENT_CURRENT_PREFIX:="/foo/bar"}',
                "expected": ': ${AMENT_CURRENT_PREFIX:="$SNAP_COLCON_ROOT/foo/bar"}',
            },
            {
                "path": "test/no-change.sh",
                "contents": ": ${AMENT_CURRENT_PREFIX:=$NOT_A_PATH}",
                "expected": ": ${AMENT_CURRENT_PREFIX:=$NOT_A_PATH}",
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._prepare_build()

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))


class FinishBuildTest(ColconPluginTestBase):
    def setUp(self):
        super().setUp()
        self.plugin = colcon.ColconPlugin("test-part", self.properties, self.project)

    @mock.patch("snapcraft.internal.mangling.rewrite_python_shebangs")
    def test_in_snap_python_is_used(self, shebangs_mock):
        # Mangling has its own tests. Here we just need to make sure
        # _prepare_build actually uses it.
        self.plugin._finish_build()
        shebangs_mock.assert_called_once_with(self.plugin.installdir)

    def test_cmake_paths_are_rewritten(self):
        os.makedirs(os.path.join(self.plugin._ros_underlay, "test"))

        # Place a few .cmake files with incorrect paths, and some files that
        # shouldn't be changed.
        files = [
            {
                "path": "fooConfig.cmake",
                "contents": '"{}/usr/lib/foo"'.format(self.plugin.installdir),
                "expected": '"$ENV{SNAPCRAFT_STAGE}/usr/lib/foo"',
            },
            {
                "path": "bar.cmake",
                "contents": '"/usr/lib/bar"',
                "expected": '"/usr/lib/bar"',
            },
            {
                "path": "test/bazConfig.cmake",
                "contents": '"{0}/test/baz;{0}/usr/lib/baz"'.format(
                    self.plugin.installdir
                ),
                "expected": '"$ENV{SNAPCRAFT_STAGE}/test/baz;'
                '$ENV{SNAPCRAFT_STAGE}/usr/lib/baz"',
            },
            {"path": "test/quxConfig.cmake", "contents": "qux", "expected": "qux"},
            {
                "path": "test/installedConfig.cmake",
                "contents": '"$ENV{SNAPCRAFT_STAGE}/foo"',
                "expected": '"$ENV{SNAPCRAFT_STAGE}/foo"',
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._finish_build()

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

        # Verify that no sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python2.7", "sitecustomize.py"
            ),
            Not(FileExists()),
        )

    def test_ament_prefix_is_rewritten(self):
        os.makedirs(os.path.join(self.plugin._ros_underlay, "test"))

        # Place files with incorrect prefixes, and some that shouldn't be changed.
        files = [
            {
                "path": "setup.sh",
                "contents": ": ${AMENT_CURRENT_PREFIX:=/opt/ros/crystal}",
                "expected": ': ${AMENT_CURRENT_PREFIX:="$SNAP_COLCON_ROOT/opt/ros/crystal"}',
            },
            {
                "path": "test/local_setup.sh",
                "contents": ': ${{AMENT_CURRENT_PREFIX:="{}/foo/bar"}}'.format(
                    self.plugin.installdir
                ),
                "expected": ': ${AMENT_CURRENT_PREFIX:="$SNAP_COLCON_ROOT/foo/bar"}',
            },
            {
                "path": "test/no-change.sh",
                "contents": ": ${AMENT_CURRENT_PREFIX:=$NOT_A_PATH}",
                "expected": ": ${AMENT_CURRENT_PREFIX:=$NOT_A_PATH}",
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._finish_build()

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    def test_colcon_prefix_is_rewritten(self):
        os.makedirs(os.path.join(self.plugin._ros_underlay, "test"))

        # Place files with incorrect prefixes, and some that shouldn't be changed.
        files = [
            {
                "path": "setup.sh",
                "contents": "_colcon_prefix_sh_COLCON_CURRENT_PREFIX=/foo/bar",
                "expected": '_colcon_prefix_sh_COLCON_CURRENT_PREFIX="$SNAP_COLCON_ROOT/foo/bar"',
            },
            {
                "path": "test/local_setup.sh",
                "contents": '_colcon_package_sh_COLCON_CURRENT_PREFIX="{}/baz"'.format(
                    self.plugin.installdir
                ),
                "expected": '_colcon_package_sh_COLCON_CURRENT_PREFIX="$SNAP_COLCON_ROOT/baz"',
            },
            {
                "path": "test/another_local_setup.sh",
                "contents": 'COLCON_CURRENT_PREFIX="{}/qux"'.format(
                    self.plugin.installdir
                ),
                "expected": 'COLCON_CURRENT_PREFIX="$SNAP_COLCON_ROOT/qux"',
            },
            {
                "path": "test/no-change1.sh",
                "contents": "_colcon_prefix_chain_sh_COLCON_CURRENT_PREFIX=$NOT_A_PATH",
                "expected": "_colcon_prefix_chain_sh_COLCON_CURRENT_PREFIX=$NOT_A_PATH",
            },
            {
                "path": "test/no-change2.sh",
                "contents": "_colcon_package_bash_COLCON_CURRENT_PREFIX=/foo/bar",
                "expected": "_colcon_package_bash_COLCON_CURRENT_PREFIX=/foo/bar",
            },
            {
                "path": "test/no-change3.sh",
                "contents": "COLCON_CURRENT_PREFIX=/foo/bar",
                "expected": "COLCON_CURRENT_PREFIX=/foo/bar",
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._finish_build()

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    def test_colcon_python_executable_is_rewritten(self):
        os.makedirs(os.path.join(self.plugin._ros_underlay, "test"))

        # Place files with incorrect prefixes, and some that shouldn't be changed.
        files = [
            {
                "path": "setup.sh",
                "contents": "_colcon_python_executable=/usr/bin/python3",
                "expected": '_colcon_python_executable="$SNAP_COLCON_ROOT/usr/bin/python3"',
            },
            {
                "path": "test/local_setup.sh",
                "contents": '_colcon_python_executable="{}/baz/python3"'.format(
                    self.plugin.installdir
                ),
                "expected": '_colcon_python_executable="$SNAP_COLCON_ROOT/baz/python3"',
            },
            {
                "path": "test/no-change.sh",
                "contents": "_colcon_python_executable=$NOT_A_PATH",
                "expected": "_colcon_python_executable=$NOT_A_PATH",
            },
        ]

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "w") as f:
                f.write(file_info["contents"])

        self.plugin._finish_build()

        for file_info in files:
            path = os.path.join(self.plugin._ros_underlay, file_info["path"])
            with open(path, "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    @mock.patch.object(colcon.ColconPlugin, "run")
    @mock.patch.object(colcon.ColconPlugin, "run_output", return_value="foo")
    def test_finish_build_python_sitecustomize(self, run_output_mock, run_mock):
        self.pip_mock.return_value.list.return_value = {"test-package"}

        # Create site.py, indicating that python3 was a stage-package
        site_py_path = os.path.join(
            self.plugin.installdir, "usr", "lib", "python3.6", "site.py"
        )
        os.makedirs(os.path.dirname(site_py_path), exist_ok=True)
        open(site_py_path, "w").close()

        # Also create python3 site-packages, indicating that pip packages were
        # installed.
        os.makedirs(
            os.path.join(self.plugin.installdir, "lib", "python3.6", "site-packages"),
            exist_ok=True,
        )

        self.plugin._finish_build()

        # Verify that sitecustomize.py was generated
        self.assertThat(
            os.path.join(
                self.plugin.installdir, "usr", "lib", "python3.6", "sitecustomize.py"
            ),
            FileExists(),
        )


class PullTestCase(ColconPluginTestBase):
    def test_pull_debian_dependencies(self):
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"apt": {"foo", "bar", "baz"}}

        plugin.pull()

        self.assert_rosdep_setup(
            plugin.options.colcon_rosdistro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
            plugin.project._get_build_base(),
            plugin.project._get_stage_packages_target_arch(),
        )

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(self.dependencies_mock.call_args[0][0], Equals({"my_package"}))

        # Verify that the dependencies were installed
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        package_names={"bar", "baz", "foo"},
                        base=plugin.project._get_build_base(),
                        target_arch=plugin.project._get_stage_packages_target_arch(),
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=plugin.stage_packages_path,
                        install_path=pathlib.Path(plugin.installdir),
                    )
                ]
            ),
        )

    def test_pull_local_dependencies(self):
        self.properties.colcon_packages.append("package_2")
        self.properties.source_subdir = "subdir"

        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "subdir", "src"))

        # No system dependencies (only local)
        self.dependencies_mock.return_value = {}

        plugin.pull()

        self.assert_rosdep_setup(
            plugin.options.colcon_rosdistro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "subdir", "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
            plugin.project._get_build_base(),
            plugin.project._get_stage_packages_target_arch(),
        )

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(
            self.dependencies_mock.call_args[0][0], Equals({"my_package", "package_2"})
        )

        # Verify that no .deb packages were installed
        self.assertTrue(
            mock.call().unpack(plugin.installdir) not in self.ubuntu_mock.mock_calls
        )

    def test_pull_pip_dependencies(self):
        plugin = colcon.ColconPlugin("test-part", self.properties, self.project)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        self.dependencies_mock.return_value = {"pip": {"foo", "bar", "baz"}}

        plugin.pull()

        self.assert_rosdep_setup(
            plugin.options.colcon_rosdistro,
            self.ros_version,
            os.path.join(plugin.sourcedir, "src"),
            os.path.join(plugin.partdir, "rosdep"),
            self.ubuntu_distro,
            plugin.project._get_build_base(),
            plugin.project._get_stage_packages_target_arch(),
        )

        self.assert_pip_setup(
            "3", plugin.partdir, plugin.installdir, plugin.project.stage_dir
        )

        # Verify that dependencies were found as expected. TODO: Would really
        # like to use ANY here instead of verifying explicit arguments, but
        # Python issue #25195 won't let me.
        self.assertThat(self.dependencies_mock.call_count, Equals(1))
        self.assertThat(self.dependencies_mock.call_args[0][0], Equals({"my_package"}))

        # Verify that the pip dependencies were installed
        self.pip_mock.return_value.download.assert_called_once_with(
            {"foo", "bar", "baz"}
        )
        self.pip_mock.return_value.install.assert_called_once_with(
            {"foo", "bar", "baz"}
        )


@pytest.fixture
def options():
    class Options:
        colcon_rosdistro = "crystal"
        colcon_packages = ["my_package"]
        colcon_packages_ignore = []
        colcon_source_space = "src"
        source_subdir = None
        colcon_cmake_args = []
        colcon_catkin_cmake_args = []
        colcon_ament_cmake_args = []
        build_attributes = []
        disable_parallel = False

    return Options()


class TestBuildArgs:

    package_scenarios = [
        ("one package", {"colcon_packages": ["my_package"]}),
        ("no packages", {"colcon_packages": []}),
        ("all packages", {"colcon_packages": None}),
    ]

    build_type_scenarios = [
        ("release", {"build_attributes": []}),
        ("debug", {"build_attributes": ["debug"]}),
    ]

    cmake_args_scenarios = [
        ("with colcon-cmake-args", {"colcon_cmake_args": ["-DCMAKE"]}),
        ("without colcon-cmake-args", {"colcon_cmake_args": []}),
    ]

    catkin_args_scenarios = [
        ("with catkin-cmake-args", {"colcon_catkin_cmake_args": ["-DCATKIN"]}),
        ("without catkin-cmake-args", {"colcon_catkin_cmake_args": []}),
    ]

    ament_args_scenarios = [
        ("with ament-cmake-args", {"colcon_ament_cmake_args": ["-DAMENT"]}),
        ("without ament-cmake-args", {"colcon_ament_cmake_args": []}),
    ]

    packages_ignore_scenarios = [
        ("with packages-ignore", {"colcon_packages_ignore": ["my_package"]}),
        ("without packages-ignore", {"colcon_packages_ignore": []}),
    ]

    parallel_scenarios = (
        ("disabled", dict(disable_parallel=True)),
        ("enabled", dict(disable_parallel=False)),
    )
    scenarios = multiply_scenarios(
        package_scenarios,
        build_type_scenarios,
        cmake_args_scenarios,
        catkin_args_scenarios,
        ament_args_scenarios,
        packages_ignore_scenarios,
        parallel_scenarios,
    )

    @mock.patch.object(colcon.ColconPlugin, "run")
    @mock.patch.object(colcon.ColconPlugin, "run_output", return_value="foo")
    @mock.patch.object(colcon.ColconPlugin, "_prepare_build")
    @mock.patch.object(colcon.ColconPlugin, "_finish_build")
    def test_build(
        self,
        finish_build_mock,
        prepare_build_mock,
        run_output_mock,
        run_mock,
        tmp_work_path,
        project_core18,
        options,
        disable_parallel,
        build_attributes,
        colcon_packages_ignore,
        colcon_ament_cmake_args,
        colcon_catkin_cmake_args,
        colcon_cmake_args,
        colcon_packages,
    ):
        options.colcon_packages = colcon_packages
        options.colcon_packages_ignore = colcon_packages_ignore
        options.build_attributes.extend(build_attributes)
        options.colcon_cmake_args = colcon_cmake_args
        options.colcon_catkin_cmake_args = colcon_catkin_cmake_args
        options.colcon_ament_cmake_args = colcon_ament_cmake_args
        options.disable_parallel = disable_parallel

        plugin = colcon.ColconPlugin("test-part", options, project_core18)
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        plugin.build()

        prepare_build_mock.assert_called_once_with()

        if "debug" in build_attributes:
            build_type = "Debug"
        else:
            build_type = "Release"

        expected_command = ["colcon", "build", "--merge-install"]
        if colcon_packages:
            expected_command += ["--packages-select"] + colcon_packages
        if colcon_packages_ignore:
            expected_command += ["--packages-ignore"] + colcon_packages_ignore
        expected_command += [
            "--build-base",
            plugin.builddir,
            "--base-paths",
            f"{plugin.sourcedir}/src",
            "--install-base",
            f"{plugin.installdir}/opt/ros/snap",
            "--parallel-workers={}".format(1 if disable_parallel else 2),
            "--cmake-args",
            f"-DCMAKE_BUILD_TYPE={build_type}",
        ]
        if colcon_cmake_args:
            expected_command += ["-DCMAKE"]
        if colcon_catkin_cmake_args:
            expected_command += ["--catkin-cmake-args", "-DCATKIN"]
        if colcon_ament_cmake_args:
            expected_command += ["--ament-cmake-args", "-DAMENT"]

        if colcon_packages or colcon_packages is None:
            assert run_mock.mock_calls[0][1][0] == expected_command

        else:
            run_mock.assert_not_called()

        finish_build_mock.assert_called_once_with()


class FindSystemDependenciesTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.rosdep_mock = mock.MagicMock()
        self.rosdep_mock.get_dependencies.return_value = {"bar"}

    def test_find_system_dependencies_system_only(self):
        self.rosdep_mock.resolve_dependency.return_value = {"apt": {"baz"}}

        self.assertThat(
            colcon._find_system_dependencies({"foo"}, self.rosdep_mock),
            Equals({"apt": {"baz"}}),
        )

        self.rosdep_mock.get_dependencies.assert_called_once_with("foo")
        self.rosdep_mock.resolve_dependency.assert_called_once_with("bar")

    def test_find_system_dependencies_system_only_no_packages(self):
        self.rosdep_mock.resolve_dependency.return_value = {"apt": {"baz"}}

        self.assertThat(
            colcon._find_system_dependencies(None, self.rosdep_mock),
            Equals({"apt": {"baz"}}),
        )

        self.rosdep_mock.get_dependencies.assert_called_once_with()
        self.rosdep_mock.resolve_dependency.assert_called_once_with("bar")

    def test_find_system_dependencies_local_only(self):
        self.assertThat(
            colcon._find_system_dependencies({"foo", "bar"}, self.rosdep_mock),
            HasLength(0),
        )

        self.rosdep_mock.get_dependencies.assert_has_calls(
            [mock.call("foo"), mock.call("bar")], any_order=True
        )
        self.rosdep_mock.resolve_dependency.assert_not_called()

    def test_find_system_dependencies_missing_local_dependency(self):
        # Setup a dependency on a non-existing package, and it doesn't resolve
        # to a system dependency.'
        exception = _ros.rosdep.RosdepDependencyNotResolvedError("foo")
        self.rosdep_mock.resolve_dependency.side_effect = exception

        raised = self.assertRaises(
            colcon.ColconInvalidSystemDependencyError,
            colcon._find_system_dependencies,
            {"foo"},
            self.rosdep_mock,
        )

        self.assertThat(
            str(raised),
            Equals(
                "Package 'bar' isn't a valid system dependency. Did "
                "you forget to add it to colcon-packages? If not, "
                "add the Ubuntu package containing it to "
                "stage-packages until you can get it into the rosdep "
                "database."
            ),
        )

    def test_find_system_dependencies_raises_if_unsupported_type(self):
        self.rosdep_mock.resolve_dependency.return_value = {"unsupported-type": {"baz"}}

        raised = self.assertRaises(
            colcon.ColconUnsupportedDependencyTypeError,
            colcon._find_system_dependencies,
            {"foo"},
            self.rosdep_mock,
        )

        self.assertThat(
            str(raised),
            Equals(
                "Package 'bar' resolved to an unsupported type of dependency: "
                "'unsupported-type'."
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import pathlib
import subprocess

from snapcraft import file_utils

from . import errors

_REVIEW_TOOLS_PATH = pathlib.Path("/snap/bin/review-tools.snap-review")
_REVIEW_TOOLS_SNAP_USER_COMMON = pathlib.Path("~/snap/review-tools/common").expanduser()


def is_available() -> bool:
    return _REVIEW_TOOLS_PATH.exists()


def _get_review_tools_user_common() -> pathlib.Path:
    return _REVIEW_TOOLS_SNAP_USER_COMMON


def _snap_in_review_tools_common(method):
    def common_decorator(*, snap_filename: str) -> None:
        snap_filename_common_path = (
            _get_review_tools_user_common() / pathlib.Path(snap_filename).name
        )
        # Create the directory tree for the cases where the review-tools have not run before.
        snap_filename_common_path.parent.mkdir(parents=True, exist_ok=True)
        file_utils.link_or_copy(snap_filename, str(snap_filename_common_path))

        try:
            method(snap_filename=snap_filename_common_path)
        finally:
            if snap_filename_common_path.exists():
                snap_filename_common_path.unlink()

    return common_decorator


@_snap_in_review_tools_common
def run(*, snap_filename: str) -> None:
    # TODO allow suppression of error and warn through project configuration.
    command = [
        _REVIEW_TOOLS_PATH.as_posix(),
        snap_filename,
        "--json",
        "--allow-classic",
    ]
    # Speed up the process by not doing the re-squash tests.
    env = dict(SNAP_ENFORCE_RESQUASHFS="0")

    try:
        subprocess.check_output(command, stderr=subprocess.STDOUT, env=env)
    except FileNotFoundError:
        raise errors.ReviewToolMissing()
    except subprocess.CalledProcessError as call_error:
        # Error codes:
        # 2 -> warnings.
        # 3 -> warnings and errors.
        # Any of these trigger a store review.
        if call_error.returncode not in (2, 3):
            raise
        # We only really care about v2_lint now.
        raise errors.ReviewError(review_json=json.loads(call_error.output.decode()))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
from unittest import mock

import fixtures
from testtools.matchers import Equals

from snapcraft.internal import sources
from tests import unit


# LP: #1733584
class TestMercurial(unit.sources.SourceTestCase):  # type: ignore
    def setUp(self):
        super().setUp()
        patcher = mock.patch("snapcraft.sources.Mercurial._get_source_details")
        self.mock_get_source_details = patcher.start()
        self.mock_get_source_details.return_value = ""
        self.addCleanup(patcher.stop)

    def test_pull(self):
        hg = sources.Mercurial("hg://my-source", "source_dir")
        hg.pull()

        self.mock_run.assert_called_once_with(
            ["hg", "clone", "hg://my-source", "source_dir"]
        )

    def test_pull_branch(self):
        hg = sources.Mercurial(
            "hg://my-source", "source_dir", source_branch="my-branch"
        )
        hg.pull()

        self.mock_run.assert_called_once_with(
            ["hg", "clone", "-u", "my-branch", "hg://my-source", "source_dir"]
        )

    def test_pull_tag(self):
        hg = sources.Mercurial("hg://my-source", "source_dir", source_tag="tag")
        hg.pull()

        self.mock_run.assert_called_once_with(
            ["hg", "clone", "-u", "tag", "hg://my-source", "source_dir"]
        )

    def test_pull_commit(self):
        hg = sources.Mercurial("hg://my-source", "source_dir", source_commit="2")
        hg.pull()

        self.mock_run.assert_called_once_with(
            ["hg", "clone", "-u", "2", "hg://my-source", "source_dir"]
        )

    def test_pull_existing(self):
        self.mock_path_exists.return_value = True

        hg = sources.Mercurial("hg://my-source", "source_dir")
        hg.pull()

        self.mock_run.assert_called_once_with(["hg", "pull", "hg://my-source"])

    def test_pull_existing_with_tag(self):
        self.mock_path_exists.return_value = True

        hg = sources.Mercurial("hg://my-source", "source_dir", source_tag="tag")
        hg.pull()

        self.mock_run.assert_called_once_with(
            ["hg", "pull", "-r", "tag", "hg://my-source"]
        )

    def test_pull_existing_with_commit(self):
        self.mock_path_exists.return_value = True

        hg = sources.Mercurial("hg://my-source", "source_dir", source_commit="2")
        hg.pull()

        self.mock_run.assert_called_once_with(
            ["hg", "pull", "-r", "2", "hg://my-source"]
        )

    def test_pull_existing_with_branch(self):
        self.mock_path_exists.return_value = True

        hg = sources.Mercurial(
            "hg://my-source", "source_dir", source_branch="my-branch"
        )
        hg.pull()

        self.mock_run.assert_called_once_with(
            ["hg", "pull", "-b", "my-branch", "hg://my-source"]
        )

    def test_init_with_source_branch_and_tag_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceIncompatibleOptionsError,
            sources.Mercurial,
            "hg://mysource",
            "source_dir",
            source_tag="tag",
            source_branch="branch",
        )

        self.assertThat(raised.source_type, Equals("mercurial"))
        self.assertThat(raised.options, Equals(["source-tag", "source-branch"]))

    def test_init_with_source_commit_and_tag_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceIncompatibleOptionsError,
            sources.Mercurial,
            "hg://mysource",
            "source_dir",
            source_commit="2",
            source_tag="tag",
        )

        self.assertThat(raised.source_type, Equals("mercurial"))
        self.assertThat(raised.options, Equals(["source-tag", "source-commit"]))

    def test_init_with_source_commit_and_branch_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceIncompatibleOptionsError,
            sources.Mercurial,
            "hg://mysource",
            "source_dir",
            source_commit="2",
            source_branch="branch",
        )

        self.assertThat(raised.source_type, Equals("mercurial"))
        self.assertThat(raised.options, Equals(["source-branch", "source-commit"]))

    def test_init_with_source_depth_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceInvalidOptionError,
            sources.Mercurial,
            "hg://mysource",
            "source_dir",
            source_depth=2,
        )

        self.assertThat(raised.source_type, Equals("mercurial"))
        self.assertThat(raised.option, Equals("source-depth"))

    def test_source_checksum_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceInvalidOptionError,
            sources.Mercurial,
            "hg://mysource",
            "source_dir",
            source_checksum="md5/d9210476aac5f367b14e513bdefdee08",
        )

        self.assertThat(raised.source_type, Equals("mercurial"))
        self.assertThat(raised.option, Equals("source-checksum"))

    def test_has_source_handler_entry(self):
        self.assertTrue(sources._source_handler["mercurial"] is sources.Mercurial)

    def test_pull_failure(self):
        self.mock_run.side_effect = subprocess.CalledProcessError(1, [])

        hg = sources.Mercurial("hg://my-source", "source_dir")
        raised = self.assertRaises(sources.errors.SnapcraftPullError, hg.pull)
        self.assertThat(raised.command, Equals("hg clone hg://my-source source_dir"))
        self.assertThat(raised.exit_code, Equals(1))


def get_side_effect(original_call):
    def side_effect(cmd, *args, **kwargs):
        if len(cmd) > 1 and cmd[1] == "id":
            return "mock-commit".encode()
        elif cmd[0] == "hg":
            return
        return original_call(cmd, *args, **kwargs)

    return side_effect


class MercurialDetailsTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.working_tree = "hg-test"
        self.source_dir = "hg-checkout"

        os.mkdir(self.source_dir)
        # Simulate that we have already cloned the code.
        os.mkdir(os.path.join(self.source_dir, ".hg"))

        self.fake_check_output = self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_output",
                side_effect=get_side_effect(subprocess.check_output),
            )
        )
        self.fake_check_call = self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_call",
                side_effect=get_side_effect(subprocess.check_call),
            )
        )

    def test_hg_details_commit(self):
        hg = sources.Mercurial(self.working_tree, self.source_dir, silent=True)
        hg.pull()

        source_details = hg._get_source_details()
        self.assertThat(source_details["source-commit"], Equals("mock-commit"))

        self.fake_check_output.mock.assert_has_calls(
            [
                mock.call(["hg", "id", self.source_dir]),
                mock.call(["hg", "id", self.source_dir]),
            ]
        )
        self.fake_check_call.mock.assert_called_once_with(
            ["hg", "pull", self.working_tree], stderr=-3, stdout=-3
        )

    def test_hg_details_branch(self):
        hg = sources.Mercurial(
            self.working_tree, self.source_dir, silent=True, source_branch="test-branch"
        )
        hg.pull()

        source_details = hg._get_source_details()
        self.assertThat(source_details["source-branch"], Equals("test-branch"))

        self.fake_check_output.mock.assert_not_called()
        self.fake_check_call.mock.assert_called_once_with(
            ["hg", "pull", "-b", "test-branch", self.working_tree], stderr=-3, stdout=-3
        )

    def test_hg_details_tag(self):
        hg = sources.Mercurial(
            self.working_tree, self.source_dir, silent=True, source_tag="test-tag"
        )
        hg.pull()

        source_details = hg._get_source_details()
        self.assertThat(source_details["source-tag"], Equals("test-tag"))

        self.fake_check_output.mock.assert_not_called()
        self.fake_check_call.mock.assert_called_once_with(
            ["hg", "pull", "-r", "test-tag", self.working_tree], stderr=-3, stdout=-3
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import pytest

from snapcraft import file_utils
from snapcraft.internal import errors
from snapcraft.internal.pluginhandler import PartPatcher
from tests.unit import load_part


@pytest.fixture
def mock_elf_patcher():
    """Return a mock for snapcraft.internal.elf.Patcher."""
    patcher = mock.patch("snapcraft.internal.elf.Patcher", autospec=True)
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def mock_partpatcher():
    """Return a mock for snapcraft.internal.pluginhandler.PartPatcher."""
    patcher = mock.patch("snapcraft.internal.pluginhandler.PartPatcher", autospec=True)
    yield patcher.start()
    patcher.stop()


class TestStaticBasePatching:
    scenarios = (
        ("strict", dict(confinement="strict")),
        ("classic", dict(confinement="classic")),
    )

    def test_static_base_with_libc6_stage_packaged(self, mock_elf_patcher, confinement):
        # The "bare" base is a static base.
        handler = load_part(
            "test-part",
            snap_type="app",
            base="bare",
            build_base="core",
            confinement=confinement,
        )

        patcher = PartPatcher(
            elf_files=frozenset(["foo"]),
            project=handler._project,
            snap_base_path="/snap/test-snap/current",
            stage_packages=["libc6"],
        )

        patcher.patch()

        mock_elf_patcher.assert_called_once_with(
            dynamic_linker="/snap/test-snap/current/lib/x86_64-linux-gnu/ld-2.27.so",
            preferred_patchelf_path=None,
            root_path=handler._project.prime_dir,
        )

    def test_static_base_without_libc6_stage_packaged(
        self, mock_elf_patcher, confinement
    ):
        # The "bare" base is a static base, empty, so there is no linker loader to look for.
        handler = load_part(
            "test-part",
            snap_type="app",
            base="bare",
            build_base="core",
            confinement=confinement,
        )

        patcher = PartPatcher(
            elf_files=frozenset(["foo"]),
            project=handler._project,
            snap_base_path="/snap/test-snap/current",
            stage_packages=[],
        )

        patcher.patch()

        mock_elf_patcher.assert_not_called()

    def test_no_base(self, mock_elf_patcher, confinement):
        handler = load_part(
            "test-part",
            snap_type="app",
            base=None,
            build_base="core",
            confinement=confinement,
        )

        patcher = PartPatcher(
            elf_files=frozenset(["foo"]),
            project=handler._project,
            snap_base_path="/snap/test-snap/current",
            stage_packages=[],
        )

        patcher.patch()

        mock_elf_patcher.assert_not_called()

    def test_conflicting_patchelf_build_attributes(
        self, monkeypatch, mock_partpatcher, confinement
    ):
        monkeypatch.setattr(file_utils, "get_snap_tool_path", lambda x: x)

        part_properties = {
            "source-subdir": "src",
            "build-attributes": ["no-patchelf", "enable-patchelf"],
        }

        handler = load_part(
            "test-part",
            snap_name="test-snap",
            part_properties=part_properties,
            confinement=confinement,
        )

        with pytest.raises(errors.BuildAttributePatchelfConflictError):
            handler.prime()


class TestPrimeTypeExcludesPatching:

    scenarios = (
        (
            "kernel",
            dict(
                snap_type="kernel",
                snap_name="test-snap",
                confinement=None,
                build_attributes=None,
            ),
        ),
        (
            "gadget",
            dict(
                snap_type="gadget",
                snap_name="test-snap",
                confinement=None,
                build_attributes=None,
            ),
        ),
        (
            "base",
            dict(
                snap_type="base",
                snap_name="core18",
                confinement=None,
                build_attributes=None,
            ),
        ),
        (
            "os",
            dict(
                snap_type="os",
                snap_name="test-snap",
                confinement=None,
                build_attributes=None,
            ),
        ),
        (
            "strict app",
            dict(
                snap_type="app",
                snap_name="test-snap",
                confinement="strict",
                build_attributes=None,
            ),
        ),
        (
            "classic but disabled",
            dict(
                snap_type="app",
                snap_name="test-snap",
                confinement="classic",
                build_attributes=["no-patchelf"],
            ),
        ),
    )

    def test_no_patcher_called(
        self,
        monkeypatch,
        mock_partpatcher,
        snap_type,
        snap_name,
        confinement,
        build_attributes,
    ):
        monkeypatch.setattr(file_utils, "get_snap_tool_path", lambda x: x)

        part_properties = {"source-subdir": "src"}
        if build_attributes:
            part_properties["build-attributes"] = build_attributes

        handler = load_part(
            "test-part",
            snap_name=snap_name,
            part_properties=part_properties,
            snap_type=snap_type,
            confinement=confinement,
        )

        handler.prime()

        mock_partpatcher.assert_not_called()


class TestPrimeTypeIncludesPatching:

    scenarios = (
        (
            "classic",
            dict(confinement="classic", build_attributes=None, stage_packages=None),
        ),
        (
            "strict but enabled",
            dict(
                confinement="strict",
                build_attributes=["enable-patchelf"],
                stage_packages=None,
            ),
        ),
        (
            "libc staged",
            dict(confinement="strict", build_attributes=None, stage_packages=["libc6"]),
        ),
    )

    def test_patcher_called(
        self,
        monkeypatch,
        mock_partpatcher,
        confinement,
        build_attributes,
        stage_packages,
    ):
        monkeypatch.setattr(file_utils, "get_snap_tool_path", lambda x: x)

        part_properties = {"source-subdir": "src"}
        if build_attributes:
            part_properties["build-attributes"] = build_attributes
        if stage_packages:
            part_properties["stage-packages"] = stage_packages
        else:
            stage_packages = list()

        handler = load_part(
            "test-part",
            snap_type="app",
            part_properties=part_properties,
            confinement=confinement,
        )

        handler.prime()

        mock_partpatcher.assert_called_with(
            elf_files=frozenset(),
            project=mock.ANY,
            snap_base_path="/snap/fake-name/current",
            stage_packages=stage_packages,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Handle surfacing deprecation notices.

When a new deprecation has occurred, write a Deprecation Notice for it on the
wiki, assigning it the next ID (DN1, DN2, etc.). Then add that ID along with a
brief message for surfacing to the user into the _DEPRECATION_MESSAGES in this
module.
"""

import logging

_DEPRECATION_MESSAGES = {
    "dn1": "The 'snap' keyword has been replaced by 'prime'.",
    "dn2": "Custom plugins should now be placed in 'snap/plugins'.",
    "dn3": "Assets in 'setup/gui' should now be placed in 'snap/gui'.",
    "dn4": "The 'history' command has been replaced by 'list-revisions'.",
    "dn5": "Aliases are now handled by the store, and shouldn't be declared "
    "in the snap.",
    "dn6": "Use of the 'snap' command with a directory has been deprecated "
    "in favour of the 'pack' command.",
    "dn7": "The 'prepare' keyword has been replaced by 'override-build'",
    "dn8": "The 'build' keyword has been replaced by 'override-build'",
    "dn9": "The 'install' keyword has been replaced by 'override-build'",
    "dn10": "The 'version-script' keyword has been deprecated in favour of "
    "the 'snapcraftctl set-version' part scriptlet.",
    "dn11": "The 'push' set of commands have been replaced with 'upload'.",
    "dn12": "The 'registered' and 'list-registered' commands have with replaced with 'list'.",
    "dn13": "Support will be removed for legacy and core-based projects in Snapcraft 5.0 (expected July 22, 2021).",
}

_DEPRECATION_URL_FMT = "http://snapcraft.io/docs/deprecation-notices/{id}"

logger = logging.getLogger(__name__)


def _deprecation_message(id):
    message = _DEPRECATION_MESSAGES.get(id)
    if not message:
        raise RuntimeError("No deprecation notice with id {!r}".format(id))

    return message


def handle_deprecation_notice(id):
    message = _deprecation_message(id)

    logger.warning(
        "DEPRECATED: {}\nSee {} for more information.".format(
            message, _DEPRECATION_URL_FMT.format(id=id)
        )
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The catkin tools plugin for ROS 1 parts.

    - catkin-tools-packages:
      (list of strings)
      List of catkin packages to build. If not specified, all packages in the
      workspace will be built. If set to an empty list ([]), no packages will
      be built, which could be useful if you only want ROS debs in the snap.

    - catkin-tools-cmake-args:
      (list of strings)
      Arguments to pass to cmake projects.

    This plugin requires certain variables that are specified by the `ros1-noetic`
    extension. If you're not using the extension, set these in your `build-environment`:
      - ROS_DISTRO: "noetic"
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import _ros


class CatkinToolsPlugin(_ros.RosPlugin):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "catkin-tools-cmake-args": {
                    "type": "array",
                    "minItems": 0,
                    "items": {"type": "string"},
                    "default": [],
                },
                "catkin-tools-packages": {
                    "type": "array",
                    "minItems": 0,
                    "uniqueItems": True,
                    "items": {"type": "string"},
                },
            },
        }

    def get_build_packages(self) -> Set[str]:
        return super().get_build_packages() | {
            "python3-catkin-tools",
            # FIXME: Only needed because of a botched release:
            # https://github.com/catkin/catkin_tools/issues/594#issuecomment-688149976
            # Once fixed, remove this.
            "python3-osrf-pycommon",
        }

    def _get_build_commands(self) -> List[str]:
        # It's possible that this workspace wasn't initialized to be used with
        # catkin-tools, so initialize it first. Note that this is a noop if it
        # was already initialized
        commands = ["catkin init"]

        # Overwrite the default catkin profile to ensure builds
        # aren't affected by profile changes
        commands.append("catkin profile add -f default")

        # Use catkin config to set configurations for the snap build
        catkin_config_command = [
            "catkin",
            "config",
            "--profile",
            "default",
            "--install",
            "--source-space",
            "$SNAPCRAFT_PART_SRC",
            "--build-space",
            "$SNAPCRAFT_PART_BUILD",
            "--install-space",
            "$SNAPCRAFT_PART_INSTALL/opt/ros/$ROS_DISTRO",
        ]

        if self.options.catkin_tools_cmake_args:
            catkin_config_command.extend(
                ["--cmake-args", *self.options.catkin_tools_cmake_args]
            )

        commands.append(" ".join(catkin_config_command))

        # Now actually build the package
        catkin_command = [
            "catkin",
            "build",
            "--no-notify",
            "--profile",
            "default",
            "-j",
            "$SNAPCRAFT_PARALLEL_BUILD_COUNT",
        ]

        if self.options.catkin_tools_packages:
            catkin_command.extend(self.options.catkin_tools_packages)

        commands.append(" ".join(catkin_command))

        return commands

    def _get_workspace_activation_commands(self) -> List[str]:
        return [
            'state="$(set +o)"',
            "set +u",
            "_CATKIN_SETUP_DIR=/opt/ros/$ROS_DISTRO . /opt/ros/$ROS_DISTRO/setup.sh",
            'eval "$(state)"',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Equals

from . import ProjectLoaderBaseTest


class TestBuildSnapsFromSnapcraftYaml(ProjectLoaderBaseTest):
    def test_build_snaps_v1(self):
        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1.0"
            summary: test
            description: test
            confinement: strict
            grade: stable


            parts:
              part1:
                plugin: nil
                build-snaps: [foo/latest/stable, bar/default/edge]
            """
        )

        project_config = self.make_snapcraft_project(snapcraft_yaml)

        self.assertThat(
            project_config.get_build_snaps(),
            Equals({"foo/latest/stable", "core18", "bar/default/edge"}),
        )

    def test_build_snaps_v2(self):
        snapcraft_yaml = dedent(
            """\
            name: test
            base: core20
            version: "1.0"
            summary: test
            description: test
            confinement: strict
            grade: stable


            parts:
              part1:
                plugin: nil
                build-snaps: [foo/latest/stable, bar/default/edge]
            """
        )

        project_config = self.make_snapcraft_project(snapcraft_yaml)

        self.assertThat(
            project_config.get_build_snaps(),
            Equals({"foo/latest/stable", "core20", "bar/default/edge"}),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Optional

import jsonschema

from ._api_schema import BUILD_ASSERTION_JSONSCHEMA


class Snap:
    """Represent a Snap in a Validation Set."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Snap":
        jsonschema.validate(
            payload, BUILD_ASSERTION_JSONSCHEMA["properties"]["snaps"]["items"]
        )

        return cls(
            name=payload["name"],
            snap_id=payload.get("id"),
            presence=payload.get("presence"),
            revision=payload.get("revision"),
        )

    def marshal(self) -> Dict[str, Any]:
        payload = {"name": self.name}

        if self.snap_id is not None:
            payload["id"] = self.snap_id

        if self.presence is not None:
            payload["presence"] = self.presence

        if self.revision is not None:
            payload["revision"] = self.revision

        return payload

    def __eq__(self, other) -> bool:
        if not isinstance(other, Snap):
            return False

        return (
            self.name == other.name
            and self.snap_id == other.snap_id
            and self.presence == other.presence
            and self.revision == other.revision
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name!r}>"

    def __init__(
        self,
        name: str,
        snap_id: Optional[str] = None,
        presence: Optional[str] = None,
        revision: Optional[str] = None,
    ):
        self.name = name
        self.snap_id = snap_id
        self.presence = presence
        self.revision = revision


class BuildAssertion:
    """Represent Validation Set assertion headers ready local signing."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "BuildAssertion":
        jsonschema.validate(payload, BUILD_ASSERTION_JSONSCHEMA)

        return cls(
            account_id=payload["account-id"],
            authority_id=payload["authority-id"],
            name=payload["name"],
            revision=payload.get("revision", "0"),
            sequence=payload["sequence"],
            series=payload["series"],
            snaps=[Snap.unmarshal(s) for s in payload["snaps"]],
            timestamp=payload["timestamp"],
            assertion_type=payload["type"],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "account-id": self.account_id,
            "authority-id": self.authority_id,
            "name": self.name,
            "revision": self.revision,
            "sequence": self.sequence,
            "snaps": [s.marshal() for s in self.snaps],
            "timestamp": self.timestamp,
            "type": self.assertion_type,
            "series": self.series,
        }

    def __eq__(self, other) -> bool:
        if not isinstance(other, BuildAssertion):
            return False

        return (
            self.name == other.name
            and self.account_id == other.account_id
            and self.authority_id == other.authority_id
            and self.revision == other.revision
            and self.sequence == other.sequence
            and self.snaps == other.snaps
            and self.timestamp == other.timestamp
            and self.assertion_type == other.assertion_type
            and self.series == other.series
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name!r}>"

    def __init__(
        self,
        *,
        account_id: str,
        authority_id: str,
        name: str,
        revision: str,
        sequence: str,
        snaps: List[Snap],
        timestamp: str,
        assertion_type: str,
        series: str = "16",
    ):
        self.account_id = account_id
        self.authority_id = authority_id
        self.name = name
        self.revision = revision
        self.sequence = sequence
        self.snaps = snaps
        self.series = series
        self.timestamp = timestamp
        self.assertion_type = assertion_type


class ValidationSets:
    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "ValidationSets":
        return cls(
            assertions=[
                BuildAssertion.unmarshal(a["headers"])
                for a in payload.get("assertions", list())
            ]
        )

    def marshal(self) -> Dict[str, Any]:
        return {"assertions": [{"headers": a.marshal()} for a in self.assertions]}

    def __repr__(self) -> str:
        assertion_count = len(self.assertions)
        return f"<{self.__class__.__name__}: assertions {assertion_count!r}>"

    def __init__(self, assertions: List[BuildAssertion]) -> None:
        self.assertions = assertions



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

from testtools.matchers import Contains

from tests import fixture_setup, integration


class PartsGrammarTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()
        self.test_source = "https://github.com/snapcrafters/fork-and-rename-me.git"

    def test_plain_source_string(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "my-part", {"plugin": "nil", "source": self.test_source}
        )
        self.useFixture(snapcraft_yaml)
        self.run_snapcraft(["pull"])

    def test_source_on_current_arch(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "my-part",
            {
                "plugin": "nil",
                "source": [
                    {"on {}".format(self.deb_arch): self.test_source},
                    {"else": "invalid"},
                ],
            },
        )
        self.useFixture(snapcraft_yaml)
        self.run_snapcraft(["pull"])

    def test_source_try(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "my-part",
            {
                "plugin": "nil",
                "source": [{"try": self.test_source}, {"else": "invalid"}],
            },
        )
        self.useFixture(snapcraft_yaml)
        self.run_snapcraft(["pull"])

    def test_source_on_other_arch(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "my-part", {"plugin": "nil", "source": [{"on other-arch": "invalid"}]}
        )
        self.useFixture(snapcraft_yaml)
        self.run_snapcraft(["pull"])

    def test_source_on_other_arch_else_fail(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "my-part",
            {"plugin": "nil", "source": [{"on other-arch": "invalid"}, "else fail"]},
        )
        self.useFixture(snapcraft_yaml)
        self.assertThat(
            self.assertRaises(
                subprocess.CalledProcessError, self.run_snapcraft, ["pull"]
            ).output,
            Contains("Unable to satisfy 'on other-arch', failure forced"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib

from snapcraft.internal.project_loader import errors


def test_SnapcraftProjectUnusedKeyAssetError():
    path = pathlib.Path("foo")

    error = errors.SnapcraftProjectUnusedKeyAssetError(key_path=path)

    assert error.get_brief() == "Found unused key asset 'foo'."
    assert error.get_details() == "All configured key assets must be utilized."
    assert error.get_resolution() == "Verify key usage and remove all unused keys."



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import plainbox_provider
from tests import fixture_setup, unit

from . import PluginsV1BaseTestCase


class PlainboxProviderPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        """Test validity of the Scons Plugin schema"""
        schema = plainbox_provider.PlainboxProviderPlugin.schema()

        # Verify the presence of all properties
        properties = schema["properties"]
        self.assertThat(properties, Equals({}))

    def test_get_pull_properties(self):
        expected_pull_properties = []
        resulting_pull_properties = (
            plainbox_provider.PlainboxProviderPlugin.get_pull_properties()
        )

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = []
        resulting_build_properties = (
            plainbox_provider.PlainboxProviderPlugin.get_build_properties()
        )

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


class PlainboxProviderPluginTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            source = "."

        self.options = Options()

        patcher = mock.patch.object(plainbox_provider.PlainboxProviderPlugin, "run")
        self.mock_run = patcher.start()
        self.addCleanup(patcher.stop)

    def test_build(self):
        plugin = plainbox_provider.PlainboxProviderPlugin(
            "test-part", self.options, self.project
        )

        os.makedirs(plugin.sourcedir)

        # Place a few files with bad shebangs, and some files that shouldn't be
        # changed.
        files = [
            {
                "path": os.path.join(plugin.installdir, "baz"),
                "contents": "#!/foo/bar/baz/python3",
                "expected": "#!/usr/bin/env python3",
            },
            {
                "path": os.path.join(plugin.installdir, "bin", "foobar"),
                "contents": "#!/foo/baz/python3.5",
                "expected": "#!/usr/bin/env python3.5",
            },
            {
                "path": os.path.join(plugin.installdir, "foo"),
                "contents": "foo",
                "expected": "foo",
            },
            {
                "path": os.path.join(plugin.installdir, "bar"),
                "contents": "bar\n#!/usr/bin/python3",
                "expected": "bar\n#!/usr/bin/python3",
            },
        ]

        for file_info in files:
            os.makedirs(os.path.dirname(file_info["path"]), exist_ok=True)
            with open(file_info["path"], "w") as f:
                f.write(file_info["contents"])

        plugin.build()

        env = os.environ.copy()
        env["PROVIDERPATH"] = ""
        calls = [
            mock.call(["python3", "manage.py", "validate"], env=env),
            mock.call(["python3", "manage.py", "build"]),
            mock.call(["python3", "manage.py", "i18n"]),
            mock.call(
                [
                    "python3",
                    "manage.py",
                    "install",
                    "--layout=relocatable",
                    "--prefix=/providers/test-part",
                    "--root={}".format(plugin.installdir),
                ]
            ),
        ]
        self.mock_run.assert_has_calls(calls)

        for file_info in files:
            with open(os.path.join(plugin.installdir, file_info["path"]), "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    def test_build_with_provider_stage_dir(self):
        self.useFixture(fixture_setup.CleanEnvironment())

        plugin = plainbox_provider.PlainboxProviderPlugin(
            "test-part", self.options, self.project
        )

        os.makedirs(plugin.sourcedir)
        provider_path = os.path.join(
            self.project.stage_dir, "providers", "test-provider"
        )
        os.makedirs(provider_path)

        # Place a few files with bad shebangs, and some files that shouldn't be
        # changed.
        files = [
            {
                "path": os.path.join(plugin.installdir, "baz"),
                "contents": "#!/foo/bar/baz/python3",
                "expected": "#!/usr/bin/env python3",
            },
            {
                "path": os.path.join(plugin.installdir, "bin", "foobar"),
                "contents": "#!/foo/baz/python3.5",
                "expected": "#!/usr/bin/env python3.5",
            },
            {
                "path": os.path.join(plugin.installdir, "foo"),
                "contents": "foo",
                "expected": "foo",
            },
            {
                "path": os.path.join(plugin.installdir, "bar"),
                "contents": "bar\n#!/usr/bin/python3",
                "expected": "bar\n#!/usr/bin/python3",
            },
        ]

        for file_info in files:
            os.makedirs(os.path.dirname(file_info["path"]), exist_ok=True)
            with open(file_info["path"], "w") as f:
                f.write(file_info["contents"])

        plugin.build()

        calls = [
            mock.call(
                ["python3", "manage.py", "validate"],
                env={"PROVIDERPATH": provider_path},
            ),
            mock.call(["python3", "manage.py", "build"]),
            mock.call(["python3", "manage.py", "i18n"]),
            mock.call(
                [
                    "python3",
                    "manage.py",
                    "install",
                    "--layout=relocatable",
                    "--prefix=/providers/test-part",
                    "--root={}".format(plugin.installdir),
                ]
            ),
        ]
        self.mock_run.assert_has_calls(calls)

        for file_info in files:
            with open(os.path.join(plugin.installdir, file_info["path"]), "r") as f:
                self.assertThat(f.read(), Equals(file_info["expected"]))

    def test_fileset_ignores(self):
        plugin = plainbox_provider.PlainboxProviderPlugin(
            "test-part", self.options, self.project
        )

        expected_fileset = [
            "-usr/lib/python*/sitecustomize.py",
            "-etc/python*/sitecustomize.py",
        ]
        fileset = plugin.snap_fileset()
        self.assertListEqual(expected_fileset, fileset)


class PlainboxProviderPluginUnsupportedBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            plainbox_provider.PlainboxProviderPlugin,
            "test-part",
            self.options,
            self.project,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import shutil
import subprocess
from collections import OrderedDict
from unittest import mock

import pytest
from testtools.matchers import Equals

from snapcraft.internal.build_providers import errors
from snapcraft.internal.build_providers._multipass import MultipassCommand
from tests import unit


class MultipassCommandBaseTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("signal.signal")
        patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("shutil.which", return_value=["multipass"])
        self.which_mock = patcher.start()
        self.addCleanup(patcher.stop)


class MultipassCommandGeneralTest(MultipassCommandBaseTest):
    def test_provider_name(self):
        self.assertThat(MultipassCommand.provider_name, Equals("multipass"))

    def test_multipass_command_missing_raises(self):
        self.which_mock.return_value = []

        self.assertRaises(errors.ProviderNotFound, MultipassCommand, platform="linux")

    def test_ensure_multipass_on_command_found(self):
        self.which_mock.return_value = ["multipass"]

        MultipassCommand.ensure_multipass(platform="linux")


class TestMultipassCommandEnsureMultipassErrors:
    scenarios = (
        ("linux", dict(platform="linux", prompt_installable=True, snap_which=True)),
        (
            "linux not snap",
            dict(platform="linux", prompt_installable=False, snap_which=False),
        ),
        ("darwin", dict(platform="darwin", prompt_installable=True, snap_which=False)),
        (
            "unknown",
            dict(platform="unknown", prompt_installable=False, snap_which=False),
        ),
    )

    def test_ensure_multipass_raises(
        self, monkeypatch, platform, prompt_installable, snap_which
    ):
        def which_effect(command: str):
            if command == "snap" and snap_which:
                return ["snap"]

            return None

        monkeypatch.setattr(shutil, "which", which_effect)

        with pytest.raises(errors.ProviderNotFound) as error:
            MultipassCommand.ensure_multipass(platform)
            assert error.prompt_installable is prompt_installable


class MultipassCommandSetupMultipassTest(MultipassCommandBaseTest):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("time.sleep")
        patcher.start()
        self.addCleanup(patcher.stop)

        def check_output_effect(*args, **kwargs):
            if args[0] == ["multipass", "version"]:
                return b"multipassd"

        patcher = mock.patch("subprocess.check_call")
        self.check_call_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.check_output")
        self.check_output_mock = patcher.start()
        self.check_output_mock.side_effect = check_output_effect
        self.addCleanup(patcher.stop)

        self.echoer_mock = mock.Mock()

    def test_linux(self):
        self.fake_snapd.snaps_result = [
            dict(name="multipass", channel="beta", revision="10")
        ]
        self.fake_snapd.find_result = [
            dict(
                multipass=dict(
                    channel="beta",
                    type="app",
                    channels={"latest/stable": dict(confinement="classic")},
                )
            )
        ]
        MultipassCommand.setup_multipass(platform="linux", echoer=self.echoer_mock)

        self.echoer_mock.wrapped.assert_called_once_with("Waiting for multipass...")

    def test_darwin(self):
        MultipassCommand.setup_multipass(platform="darwin", echoer=self.echoer_mock)

        self.check_call_mock.assert_called_once_with(
            ["brew", "cask", "install", "multipass"]
        )
        self.echoer_mock.wrapped.assert_called_once_with("Waiting for multipass...")

    def test_unkown_platform_raises(self):
        self.assertRaises(
            EnvironmentError,
            MultipassCommand.setup_multipass,
            platform="unknown",
            echoer=self.echoer_mock,
        )

        self.echoer_mock.wrapped.assert_not_called()


class MultipassCommandPassthroughBaseTest(MultipassCommandBaseTest):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("subprocess.check_call")
        self.check_call_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.check_output")
        self.check_output_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.Popen")
        self.popen_mock = patcher.start()
        self.popen_mock.return_value.communicate.return_value = (b"", b"error")
        self.popen_mock.return_value.returncode = 0
        self.addCleanup(patcher.stop)

        self.multipass_command = MultipassCommand(platform="linux")
        self.instance_name = "stub-instance"


class MultipassCommandLaunchTest(MultipassCommandPassthroughBaseTest):
    def test_launch(self):
        self.multipass_command.launch(instance_name=self.instance_name, image="16.04")

        self.check_call_mock.assert_called_once_with(
            ["multipass", "launch", "16.04", "--name", self.instance_name],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_launch_with_cpus(self):
        self.multipass_command.launch(
            instance_name=self.instance_name, cpus="10", image="16.04"
        )

        self.check_call_mock.assert_called_once_with(
            [
                "multipass",
                "launch",
                "16.04",
                "--name",
                self.instance_name,
                "--cpus",
                "10",
            ],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_launch_with_mem(self):
        self.multipass_command.launch(
            instance_name=self.instance_name, mem="2G", image="16.04"
        )

        self.check_call_mock.assert_called_once_with(
            [
                "multipass",
                "launch",
                "16.04",
                "--name",
                self.instance_name,
                "--mem",
                "2G",
            ],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_launch_with_disk(self):
        self.multipass_command.launch(
            instance_name=self.instance_name, disk="8G", image="16.04"
        )

        self.check_call_mock.assert_called_once_with(
            [
                "multipass",
                "launch",
                "16.04",
                "--name",
                self.instance_name,
                "--disk",
                "8G",
            ],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_launch_with_remote(self):
        self.multipass_command.launch(
            instance_name=self.instance_name, image="18.04", remote="daily"
        )

        self.check_call_mock.assert_called_once_with(
            ["multipass", "launch", "daily:18.04", "--name", self.instance_name],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_launch_fails(self):
        # multipass can fail due to several reasons and will display the error
        # right above this exception message, it looks something like:
        #
        #   Creating a build environment named 'nonfilterable-mayola'
        #   failed to launch: failed to obtain exit status for remote process
        #   An error occurred when trying to launch the instance with 'multipass'.  # noqa: E501
        cmd = ["multipass", "launch", "18.04", "--name", self.instance_name]
        self.check_call_mock.side_effect = subprocess.CalledProcessError(1, cmd)

        self.assertRaises(
            errors.ProviderLaunchError,
            self.multipass_command.launch,
            instance_name=self.instance_name,
            image="18.04",
        )
        self.check_call_mock.assert_called_once_with(cmd, stdin=subprocess.DEVNULL)
        self.check_output_mock.assert_not_called()


class MultipassCommandShellTest(MultipassCommandPassthroughBaseTest):
    def test_shell(self):
        self.multipass_command.shell(instance_name=self.instance_name)

        self.check_call_mock.assert_called_once_with(
            ["multipass", "shell", self.instance_name], stdin=None
        )
        self.check_output_mock.assert_not_called()

    def test_shell_fails(self):
        # multipass can fail due to several reasons and will display the error
        # right above this exception message.
        cmd = ["multipass", "shell", self.instance_name]
        self.check_call_mock.side_effect = subprocess.CalledProcessError(1, cmd)

        self.assertRaises(
            errors.ProviderShellError,
            self.multipass_command.shell,
            instance_name=self.instance_name,
        )
        self.check_call_mock.assert_called_once_with(cmd, stdin=None)
        self.check_output_mock.assert_not_called()


class MultipassCommandStopTest(MultipassCommandPassthroughBaseTest):
    def test_stop(self):
        self.multipass_command.stop(instance_name=self.instance_name)

        self.check_call_mock.assert_called_once_with(
            ["multipass", "stop", self.instance_name], stdin=subprocess.DEVNULL
        )
        self.check_output_mock.assert_not_called()

    def test_stop_time(self):
        self.multipass_command.stop(instance_name=self.instance_name, time=10)

        self.check_call_mock.assert_called_once_with(
            ["multipass", "stop", "--time", "10", self.instance_name],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_stop_fails(self):
        # multipass can fail due to several reasons and will display the error
        # right above this exception message.
        cmd = ["multipass", "stop", self.instance_name]
        self.check_call_mock.side_effect = subprocess.CalledProcessError(1, cmd)

        self.assertRaises(
            errors.ProviderStopError,
            self.multipass_command.stop,
            instance_name=self.instance_name,
        )
        self.check_call_mock.assert_called_once_with(cmd, stdin=subprocess.DEVNULL)
        self.check_output_mock.assert_not_called()


class MultipassCommandDeleteTest(MultipassCommandPassthroughBaseTest):
    def test_delete_implicit_purge(self):
        self.multipass_command.delete(instance_name=self.instance_name)

        self.check_call_mock.assert_called_once_with(
            ["multipass", "delete", self.instance_name, "--purge"],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_delete_no_purge(self):
        self.multipass_command.delete(instance_name=self.instance_name, purge=False)

        self.check_call_mock.assert_called_once_with(
            ["multipass", "delete", self.instance_name], stdin=subprocess.DEVNULL
        )
        self.check_output_mock.assert_not_called()

    def test_delete_fails(self):
        # multipass can fail due to several reasons and will display the error
        # right above this exception message.
        cmd = ["multipass", "delete", self.instance_name, "--purge"]
        self.check_call_mock.side_effect = subprocess.CalledProcessError(1, cmd)

        self.assertRaises(
            errors.ProviderDeleteError,
            self.multipass_command.delete,
            instance_name=self.instance_name,
        )
        self.check_call_mock.assert_called_once_with(cmd, stdin=subprocess.DEVNULL)
        self.check_output_mock.assert_not_called()


class MultipassCommandMountTest(MultipassCommandPassthroughBaseTest):
    def test_mount(self):
        source = "mountpath"
        target = "{}/mountpoint".format(self.instance_name)
        self.multipass_command.mount(source=source, target=target)

        self.check_call_mock.assert_called_once_with(
            ["multipass", "mount", source, target], stdin=subprocess.DEVNULL
        )
        self.check_output_mock.assert_not_called()

    def test_mount_uid(self):
        source = "mountpath"
        target = "{}/mountpoint".format(self.instance_name)
        self.multipass_command.mount(
            source=source,
            target=target,
            uid_map=OrderedDict([("1000", "0"), ("900", "0")]),
        )

        self.check_call_mock.assert_called_once_with(
            [
                "multipass",
                "mount",
                source,
                target,
                "--uid-map",
                "1000:0",
                "--uid-map",
                "900:0",
            ],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_mount_gid(self):
        source = "mountpath"
        target = "{}/mountpoint".format(self.instance_name)
        self.multipass_command.mount(
            source=source,
            target=target,
            gid_map=OrderedDict([("1000", "0"), ("900", "0")]),
        )

        self.check_call_mock.assert_called_once_with(
            [
                "multipass",
                "mount",
                source,
                target,
                "--gid-map",
                "1000:0",
                "--gid-map",
                "900:0",
            ],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_mount_uid_and_gid(self):
        source = "mountpath"
        target = "{}/mountpoint".format(self.instance_name)
        self.multipass_command.mount(
            source=source,
            target=target,
            uid_map=OrderedDict([("1000", "0"), ("900", "1")]),
            gid_map=OrderedDict([("1000", "0"), ("900", "2")]),
        )

        self.check_call_mock.assert_called_once_with(
            [
                "multipass",
                "mount",
                source,
                target,
                "--uid-map",
                "1000:0",
                "--uid-map",
                "900:1",
                "--gid-map",
                "1000:0",
                "--gid-map",
                "900:2",
            ],
            stdin=subprocess.DEVNULL,
        )
        self.check_output_mock.assert_not_called()

    def test_mount_fails(self):
        source = "mountpath"
        target = "{}/mountpoint".format(self.instance_name)
        cmd = ["multipass", "mount", source, target]
        self.check_call_mock.side_effect = subprocess.CalledProcessError(1, cmd)

        self.assertRaises(
            errors.ProviderMountError,
            self.multipass_command.mount,
            source=source,
            target=target,
        )
        self.check_call_mock.assert_called_once_with(cmd, stdin=subprocess.DEVNULL)
        self.check_output_mock.assert_not_called()


class MultipassCommandTransferTest(MultipassCommandPassthroughBaseTest):
    def test_push(self):
        source = mock.MagicMock(spec=io.BufferedIOBase)
        source.read.return_value = b"read data"
        destination = "{}/destination-file".format(self.instance_name)

        self.popen_mock.return_value.stdout = None
        self.popen_mock.return_value.communicate.side_effect = (
            subprocess.TimeoutExpired("foo", 1),
            (b"communicate data", b"error"),
        )

        self.multipass_command.push_file(source=source, destination=destination)

        self.assertEqual(
            [
                mock.call(
                    ["multipass", "transfer", "-", destination], stdin=subprocess.PIPE
                ),
                mock.call().stdin.write(b"read data"),
                mock.call().communicate(timeout=1),
                mock.call().communicate(timeout=1),
            ],
            self.popen_mock.mock_calls,
        )

        source.read.assert_called_once_with(1024)

    def test_buffered_push(self):
        source = mock.MagicMock(spec=io.BufferedIOBase)
        source.read.side_effect = (b"read data", b"")
        destination = "{}/destination-file".format(self.instance_name)

        self.popen_mock.return_value.stdout = None
        self.popen_mock.return_value.communicate.side_effect = (
            subprocess.TimeoutExpired("foo", 1),
            (b"communicate data", b"error"),
        )

        self.multipass_command.push_file(
            source=source, destination=destination, bufsize=9
        )

        self.assertEqual(
            [
                mock.call(
                    ["multipass", "transfer", "-", destination], stdin=subprocess.PIPE
                ),
                mock.call().stdin.write(b"read data"),
                mock.call().communicate(timeout=1),
                mock.call().communicate(timeout=1),
            ],
            self.popen_mock.mock_calls,
        )

        self.assertEqual(source.read.mock_calls, [mock.call(9), mock.call(9)])

    def test_pull(self):
        source = "source-file"
        destination = mock.MagicMock(spec=io.BufferedIOBase)

        self.popen_mock.return_value.stdin = None
        self.popen_mock.return_value.stdout.read.return_value = b"stdout data"
        self.popen_mock.return_value.communicate.side_effect = (
            subprocess.TimeoutExpired("foo", 1),
            (b"communicate data", b"error"),
        )

        self.multipass_command.pull_file(source=source, destination=destination)

        self.assertEqual(
            [
                mock.call(
                    ["multipass", "transfer", source, "-"],
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                ),
                mock.call().stdout.read(1024),
                mock.call().communicate(timeout=1),
                mock.call().communicate(timeout=1),
            ],
            self.popen_mock.mock_calls,
        )

        self.assertEqual(
            [mock.call(b"stdout data"), mock.call(b"communicate data")],
            destination.write.mock_calls,
        )

    def test_buffered_pull(self):
        source = "source-file"
        destination = mock.MagicMock(spec=io.BufferedIOBase)

        self.popen_mock.return_value.stdin = None
        self.popen_mock.return_value.stdout.read.side_effect = (
            b"full stdout data",
            b"stdout data",
        )

        self.popen_mock.return_value.communicate.side_effect = (
            subprocess.TimeoutExpired("foo", 1),
            (b"communicate data", b"error"),
        )

        self.multipass_command.pull_file(
            source=source, destination=destination, bufsize=16
        )

        self.assertEqual(
            [
                mock.call(
                    ["multipass", "transfer", source, "-"],
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                ),
                mock.call().stdout.read(16),
                mock.call().stdout.read(16),
                mock.call().communicate(timeout=1),
                mock.call().communicate(timeout=1),
            ],
            self.popen_mock.mock_calls,
        )

        self.assertEqual(
            [
                mock.call(b"full stdout data"),
                mock.call(b"stdout data"),
                mock.call(b"communicate data"),
            ],
            destination.write.mock_calls,
        )

    def test_buffered_pull_no_communicate_stdout(self):
        source = "source-file"
        destination = mock.MagicMock(spec=io.BufferedIOBase)

        self.popen_mock.return_value.stdin = None
        self.popen_mock.return_value.stdout.read.return_value = ""

        self.popen_mock.return_value.communicate.side_effect = (
            subprocess.TimeoutExpired("foo", 1),
            (b"", b"error"),
        )

        self.multipass_command.pull_file(source=source, destination=destination)

        destination.write.assert_not_called()

    def test_push_fails(self):
        # multipass can fail due to several reasons and will display the error
        # right above this exception message.
        source = mock.MagicMock(spec=io.BufferedIOBase)
        destination = "destination-file"
        cmd = ["multipass", "transfer", "-", destination]

        self.popen_mock.return_value.stdin = mock.MagicMock()
        self.popen_mock.return_value.stdout = None
        self.popen_mock.return_value.returncode = 1

        self.assertRaises(
            errors.ProviderFileCopyError,
            self.multipass_command.push_file,
            source=source,
            destination=destination,
        )
        self.popen_mock.assert_called_once_with(cmd, stdin=subprocess.PIPE)
        self.check_output_mock.assert_not_called()

    def test_pull_fails(self):
        # multipass can fail due to several reasons and will display the error
        # right above this exception message.
        source = "source-file"
        destination = mock.MagicMock(spec=io.BufferedIOBase)
        cmd = ["multipass", "transfer", source, "-"]

        self.popen_mock.return_value.stdin = None
        self.popen_mock.return_value.stdout = mock.MagicMock()
        self.popen_mock.return_value.returncode = 1

        self.assertRaises(
            errors.ProviderFileCopyError,
            self.multipass_command.pull_file,
            source=source,
            destination=destination,
        )
        self.popen_mock.assert_called_once_with(
            cmd, stdin=subprocess.DEVNULL, stdout=subprocess.PIPE
        )
        self.check_output_mock.assert_not_called()


class MultipassCommandExistsTest(MultipassCommandPassthroughBaseTest):
    def test_exists(self):
        exists = self.multipass_command.exists(instance_name=self.instance_name)

        self.check_call_mock.assert_has_calls(
            [
                mock.call(
                    ["multipass", "info", self.instance_name], stdin=subprocess.DEVNULL,
                )
            ]
        )

        assert exists is True

    def test_exists_false(self):
        self.check_call_mock.side_effect = subprocess.CalledProcessError(1, "info")

        exists = self.multipass_command.exists(instance_name=self.instance_name)

        self.check_call_mock.assert_has_calls(
            [
                mock.call(
                    ["multipass", "info", self.instance_name], stdin=subprocess.DEVNULL,
                )
            ]
        )

        assert exists is False


class MultipassCommandInfoTest(MultipassCommandPassthroughBaseTest):
    def test_info(self):
        self.multipass_command.info(instance_name=self.instance_name)

        self.popen_mock.assert_has_calls(
            [
                mock.call(
                    ["multipass", "info", self.instance_name],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            ]
        )
        self.check_output_mock.assert_not_called()
        self.check_call_mock.assert_not_called()

    def test_info_with_format(self):
        self.multipass_command.info(
            instance_name=self.instance_name, output_format="json"
        )

        self.popen_mock.assert_has_calls(
            [
                mock.call(
                    ["multipass", "info", self.instance_name, "--format", "json"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            ]
        )
        self.check_output_mock.assert_not_called()
        self.check_call_mock.assert_not_called()

    def test_info_fails(self):
        self.popen_mock().returncode = 1

        self.assertRaises(
            errors.ProviderInfoError,
            self.multipass_command.info,
            instance_name=self.instance_name,
        )
        self.popen_mock.assert_has_calls(
            [
                mock.call(
                    ["multipass", "info", self.instance_name],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            ]
        )
        self.check_output_mock.assert_not_called()
        self.check_call_mock.assert_not_called()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fixtures
from testtools.matchers import Contains, Equals

import snapcraft
from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class SetDefaultTrackCommandTestCase(FakeStoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        self.fake_metadata = fixtures.MockPatchObject(
            storeapi.StoreClient, "upload_metadata"
        )
        self.useFixture(self.fake_metadata)

    def test_set_default_track_without_snap_raises_exception(self):
        result = self.run_command(["set-default-track"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_set_default_track_without_login_must_ask(self):
        self.fake_metadata.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            None,
        ]

        result = self.run_command(
            ["set-default-track", "snap-test", "2.0"],
            input="user@example.com\nsecret\n",
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_set_default_track(self):
        result = self.run_command(["set-default-track", "snap-test", "2.0"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_metadata.mock.assert_called_once_with(
            snap_name="snap-test", metadata=dict(default_track="2.0"), force=True
        )

    def test_invalid_track_fails(self):
        mock_wrap = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.cli.echo.exit_error", wraps=snapcraft.cli.echo.exit_error
            )
        ).mock

        result = self.run_command(["set-default-track", "snap-test", "3.0"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("'2.0', 'latest'"))
        mock_wrap.assert_called_once_with(
            brief="The specified track '3.0' does not exist for 'snap-test'.",
            details="Valid tracks for 'snap-test': '2.0', 'latest'.",
            resolution="Ensure the '3.0' track exists for the 'snap-test' snap and try again.",
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
from typing import Any, Dict, List  # noqa
from urllib import parse

from tests import fake_servers


class FakeSnapdRequestHandler(fake_servers.BaseHTTPRequestHandler):

    snaps_result = []  # type: List[Dict[str, Any]]
    snap_details_func = None
    find_result = []  # type: List[Dict[str, Any]]
    find_exit_code = 200  # type: int
    _private_data = {"new_fake_snap_installed": False}

    def do_GET(self):
        parsed_url = parse.urlparse(self.path)
        if parsed_url.path == "/v2/snaps":
            self._handle_snaps()
        elif parsed_url.path.startswith("/v2/snaps/") and parsed_url.path.endswith(
            "file"
        ):
            self._handle_snap_file(parsed_url)
        elif parsed_url.path.startswith("/v2/snaps/"):
            self._handle_snap_details(parsed_url)
        elif parsed_url.path == "/v2/find":
            self._handle_find(parsed_url)
        else:
            self.wfile.write(parsed_url.path.encode())

    def _handle_snaps(self):
        status_code = self.find_exit_code
        params = self.snaps_result
        self.send_response(status_code)
        self.send_header("Content-Type", "text/application+json")
        self.end_headers()
        response = json.dumps({"result": params}).encode()
        self.wfile.write(response)

    def _handle_snap_file(self, parsed_url):
        self.send_response(200)
        self.send_header("Content-Length", len(parsed_url))
        self.send_header("Content-type", "text/plain")
        self.end_headers()
        self.wfile.write(parsed_url.encode())

    def _handle_snap_details(self, parsed_url):
        status_code = 404
        params = {"message": "not found"}
        type_ = "error"
        snap_name = parsed_url.path.split("/")[-1]
        if self.snap_details_func:
            status_code, params = self.snap_details_func(snap_name)
        else:
            for snap in self.snaps_result:
                if snap["name"] == snap_name:
                    status_code = 200
                    type_ = "sync"
                    params = {}
                    for key in (
                        "channel",
                        "revision",
                        "confinement",
                        "id",
                        "tracking-channel",
                    ):
                        if key in snap:
                            params.update({key: snap[key]})
                    break

        self.send_response(status_code)
        self.send_header("Content-Type", "text/application+json")
        self.end_headers()
        response = json.dumps({"result": params, "type": type_}).encode()

        self.wfile.write(response)

    def _handle_find(self, parsed_url):
        query = parse.parse_qs(parsed_url.query)
        snap_name = query["name"][0]
        status_code = 404
        params = {}
        for result in self.find_result:
            if snap_name in result:
                status_code = 200
                params = result[snap_name]
                break
        if snap_name == "new-fake-snap":
            status_code = 200
            params = {"channels": {"latest/stable": {"confinement": "strict"}}}
        elif snap_name in ("core16", "core18"):
            status_code = 200
            params = {
                "channels": {
                    "latest/stable": {"confinement": "strict", "revision": "10"}
                }
            }

        self.send_response(status_code)
        self.send_header("Content-Type", "text/application+json")
        self.end_headers()
        response = json.dumps({"result": [params]}).encode()
        self.wfile.write(response)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import logging
import os
import pkgutil

from snapcraft import extractors
from snapcraft.internal.errors import (
    InvalidExtractorValueError,
    MissingMetadataFileError,
    UnhandledMetadataFileTypeError,
)

logger = logging.getLogger(__name__)


def extract_metadata(
    part_name: str, file_relpath: str, workdir: str
) -> extractors.ExtractedMetadata:
    if not os.path.exists(os.path.join(workdir, file_relpath)):
        raise MissingMetadataFileError(part_name, file_relpath)

    # Iterate through each extractor module, calling the 'extract' function
    # from it. If it raises an 'UnhandledFileError' move onto the next.
    for _, module_name, _ in pkgutil.iter_modules(extractors.__path__):  # type: ignore
        # We only care about non-private modules in here
        if not module_name.startswith("_"):
            module = importlib.import_module(
                "snapcraft.extractors.{}".format(module_name)
            )

            try:
                # mypy is confused since we dynamically loaded the module. It
                # doesn't think it has an 'extract' function. Ignore.
                metadata = module.extract(file_relpath, workdir=workdir)  # type: ignore
                if not isinstance(metadata, extractors.ExtractedMetadata):
                    raise InvalidExtractorValueError(file_relpath, module_name)

                return metadata
            except extractors.UnhandledFileError:
                pass  # Try the next extractor
            except AttributeError:
                logger.warning(
                    "Extractor {!r} doesn't include the 'extract' function. "
                    "Skipping...".format(module_name)
                )

    # If we get here, no extractor was able to handle the file
    raise UnhandledMetadataFileTypeError(file_relpath)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from testtools.matchers import Equals

from snapcraft import project
from snapcraft.internal import steps
from snapcraft.internal.project_loader import inspection
from tests import unit


class LatestStepTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.part1 = self.load_part("part1")
        self.part2 = self.load_part("part2")
        self.parts = [self.part1, self.part2]
        self.project = project.Project()

        self.part1.makedirs()
        self.part2.makedirs()

    def test_latest_step(self):
        # Pretend part2 has been pulled, and then pretend part1 has been built
        self.part2.mark_pull_done()
        self.part1.mark_build_done()

        self.assertThat(
            inspection.latest_step(self.parts),
            Equals((self.part1, steps.BUILD, mock.ANY)),
        )

        # Now pretend part2 was built, and verify that it's now the latest step
        self.part2.mark_build_done()
        self.assertThat(
            inspection.latest_step(self.parts),
            Equals((self.part2, steps.BUILD, mock.ANY)),
        )

    def test_latest_step_no_steps_run(self):
        self.assertRaises(
            inspection.errors.NoStepsRunError, inspection.latest_step, self.parts
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
import tempfile
from textwrap import dedent
from unittest import mock

import fixtures
import xdg
from testtools.matchers import Equals, FileContains

import snapcraft.cli.echo
import snapcraft.internal.errors
from snapcraft.cli._errors import (
    _get_exception_exit_code,
    _is_reportable_error,
    _print_exception_message,
    exception_handler,
)
from snapcraft.internal.build_providers.errors import ProviderExecError
from tests import fixture_setup, unit


class SnapcraftTError(snapcraft.internal.errors.SnapcraftError):

    fmt = "{message}"

    def __init__(self, message):
        super().__init__(message=message)

    def get_exit_code(self):
        return 123


class SnapcraftTException(snapcraft.internal.errors.SnapcraftException):
    def __init__(self):
        self._brief = ""
        self._resolution = ""
        self._details = super().get_details()
        self._docs_url = super().get_docs_url()
        self._reportable = super().get_reportable()
        self._exit_code = super().get_exit_code()

    def get_brief(self):
        return self._brief

    def get_resolution(self):
        return self._resolution

    def get_details(self):
        return self._details

    def get_docs_url(self):
        return self._docs_url

    def get_exit_code(self):
        return self._exit_code

    def get_reportable(self):
        return self._reportable


class TestSnapcraftExceptionHandling(unit.TestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.cli._errors.echo.error")
        self.error_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_snapcraft_exception_format_all(self):
        exception = SnapcraftTException()
        exception._brief = "something's strange, in the neighborhood"
        exception._resolution = "who you gonna call? ghostbusters!!"
        exception._details = "i ain't afraid of no ghosts"
        exception._docs_url = "https://docs.snapcraft.io/the-snapcraft-format/8337"

        _print_exception_message(exception)
        self.error_mock.assert_called_once_with(
            """something's strange, in the neighborhood

Recommended resolution:
who you gonna call? ghostbusters!!

Detailed information:
i ain't afraid of no ghosts

For more information, check out:
https://docs.snapcraft.io/the-snapcraft-format/8337"""
        )

    def test_snapcraft_exception_minimal(self):
        exception = SnapcraftTException()
        exception._brief = "something's strange, in the neighborhood"
        exception._resolution = ""
        _print_exception_message(exception)
        self.error_mock.assert_called_once_with(
            """something's strange, in the neighborhood"""
        )

    def test_snapcraft_exception_minimal_with_resolution(self):
        exception = SnapcraftTException()
        exception._brief = "something's strange, in the neighborhood"
        exception._resolution = "who you gonna call? ghostbusters!!"
        _print_exception_message(exception)
        self.error_mock.assert_called_once_with(
            """something's strange, in the neighborhood

Recommended resolution:
who you gonna call? ghostbusters!!"""
        )

    def test_snapcraft_exception_minimal_with_resolution_and_url(self):
        exception = SnapcraftTException()
        exception._brief = "something's strange, in the neighborhood"
        exception._resolution = "who you gonna call? ghostbusters!!"
        exception._docs_url = "https://docs.snapcraft.io/the-snapcraft-format/8337"

        _print_exception_message(exception)
        self.error_mock.assert_called_once_with(
            """something's strange, in the neighborhood

Recommended resolution:
who you gonna call? ghostbusters!!

For more information, check out:
https://docs.snapcraft.io/the-snapcraft-format/8337"""
        )

    def test_snapcraft_exception_reportable(self):
        exception = SnapcraftTException()
        exception._brief = "something's strange, in the neighborhood"
        exc_info = (snapcraft.internal.errors.SnapcraftException, exception, None)

        # Test default (is false).
        self.assertFalse(_is_reportable_error(exc_info))

        # Test false.
        exception._reportable = False
        self.assertFalse(_is_reportable_error(exc_info))

        # Test true.
        exception._reportable = True
        self.assertTrue(_is_reportable_error(exc_info))

    def test_snapcraft_exception_exit_code(self):
        exception = SnapcraftTException()
        exception._brief = "something's strange, in the neighborhood"
        exception._resolution = "who you gonna call? ghostbusters!!"

        # Test default.
        self.assertEquals(2, _get_exception_exit_code(exception))

        # Test override.
        exception._exit_code = 50
        self.assertEquals(50, _get_exception_exit_code(exception))


class ReportableErrorTests(unit.TestCase):
    def test_keyboard_interrupt(self):
        exc_info = (KeyboardInterrupt, KeyboardInterrupt(), None)
        self.assertFalse(_is_reportable_error(exc_info))


class ErrorsBaseTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(fixture_setup.FakeSnapcraftIsASnap())

        patcher = mock.patch("sys.exit")
        self.exit_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("builtins.print")
        self.print_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.cli._errors.echo.error")
        self.error_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("traceback.print_exception")
        self.print_exception_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("sys.stdout.isatty", return_value=True)
        self.mock_isatty = patcher.start()
        self.addCleanup(patcher.stop)

    def call_handler(self, exception, debug):
        try:
            raise exception
        except Exception:
            exception_handler(*sys.exc_info(), debug=debug)

    def assert_exception_traceback_exit_1_with_debug(self):
        self.assertThat(self.error_mock.call_count, Equals(1))
        self.exit_mock.assert_called_once_with(1)

        # called twice - once for stdout, once for trace file
        self.assertThat(self.print_exception_mock.call_count, Equals(2))

    def assert_no_exception_traceback_exit_1_without_debug(self):
        self.assertThat(self.error_mock.call_count, Equals(1))
        self.exit_mock.assert_called_once_with(1)
        self.print_exception_mock.assert_not_called()

    def assert_print_exception_called_only_tracefile(self, error):
        self.print_exception_mock.assert_called_once_with(
            error, mock.ANY, mock.ANY, file=_Tracefile(self)
        )

    def assert_print_exception_called_both_stdout_and_tempfile(self, error):
        expected_calls = [
            mock.call(error, mock.ANY, mock.ANY, file=_Tracefile(self)),
            # TODO setup pytest runner for this to work.
            # mock.call(error, mock.ANY, mock.ANY, file=_Stdout(self)),
        ]

        self.print_exception_mock.assert_has_calls(expected_calls, any_order=True)


class ErrorsTestCase(ErrorsBaseTestCase):
    def setUp(self):
        super().setUp()

    @mock.patch.object(snapcraft.cli._errors, "RavenClient")
    @mock.patch("snapcraft.internal.common.is_snap", return_value=False)
    def test_handler_no_raven_traceback_non_snapcraft_exceptions_debug(
        self, is_snap_mock, raven_client_mock
    ):
        snapcraft.cli._errors.RavenClient = None
        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), True)
        except Exception:
            self.fail("Exception unexpectedly raised")

        self.error_mock.assert_called_once_with("not a SnapcraftError")
        self.exit_mock.assert_called_once_with(1)
        self.assert_print_exception_called_both_stdout_and_tempfile(RuntimeError)

    def test_handler_catches_snapcraft_exceptions_no_debug(self):
        try:
            self.call_handler(SnapcraftTError("is a SnapcraftError"), False)
        except Exception:
            self.fail("Exception unexpectedly raised")

        self.error_mock.assert_called_once_with("is a SnapcraftError")
        self.exit_mock.assert_called_once_with(123)
        self.print_exception_mock.assert_not_called()

    def test_handler_traces_snapcraft_exceptions_with_debug(self):
        try:
            self.call_handler(SnapcraftTError("is a SnapcraftError"), True)
        except Exception:
            self.fail("Exception unexpectedly raised")

        self.error_mock.assert_called_once_with("is a SnapcraftError")
        self.exit_mock.assert_called_once_with(123)
        self.assert_print_exception_called_both_stdout_and_tempfile(SnapcraftTError)


class ProviderErrorTest(ErrorsBaseTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("shutil.move")
        self.move_mock = patcher.start()
        self.addCleanup(patcher.stop)

    @mock.patch("os.path.isfile", return_value=False)
    def test_provider_error_legit(self, isfile_function):
        # Provider exception was raised in host environment, no crash file
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "host")
        )
        self._raise_exec_error()
        self.move_mock.assert_not_called()
        self.print_exception_mock.assert_not_called()

    @mock.patch("os.path.isfile", return_value=True)
    def test_provider_error_outer(self, isfile_function):
        # Provider exception was raised in host environment with crash file
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "host")
        )
        self._raise_exec_error()

        # Only moved if reportable
        self.move_mock.assert_not_called()
        self.print_exception_mock.assert_not_called()

    @mock.patch("os.path.isfile", return_value=False)
    def test_provider_error_host(self, isfile_function):
        # Some other error raised in the host environment
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "host")
        )
        self._raise_other_error()
        self.move_mock.assert_not_called()
        self.assertThat(self.print_exception_mock.call_count, Equals(1))

    @mock.patch("os.path.isfile", return_value=False)
    @mock.patch.object(snapcraft.cli._errors, "RavenClient")
    def test_provider_error_inner(self, isfile_function, raven_client_mock):
        # Error raised inside the build provider
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "managed-host")
        )
        snapcraft.cli._errors.RavenClient = "something"
        self._raise_other_error()
        self.move_mock.assert_not_called()
        self.assertThat(self.print_exception_mock.call_count, Equals(2))

    def _raise_exec_error(self):
        self.call_handler(
            ProviderExecError(provider_name="foo", command="bar", exit_code=2), False
        )

    def _raise_other_error(self):
        self.call_handler(KeyError, False)


class SendToSentryBaseTest(ErrorsBaseTestCase):
    def setUp(self):
        super().setUp()

        try:
            import raven  # noqa: F401
        except ImportError:
            self.skipTest("raven needs to be installed for this test.")

        patcher = mock.patch("snapcraft.cli.echo.prompt")
        self.prompt_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.cli._errors.RequestsHTTPTransport")
        self.raven_request_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.cli._errors.RavenClient")
        self.raven_client_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_ENABLE_ERROR_REPORTING", "yes")
        )


class SendToSentryFails(SendToSentryBaseTest):
    def test_send_fails(self):
        self.prompt_mock.return_value = "yes"
        self.mock_isatty.return_value = True

        self.raven_client_mock().captureException.side_effect = RuntimeError(
            "raven is broken"
        )
        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), True)
        except Exception:
            self.fail("Exception unexpectedly raised")


class SendToSentryIsYesTest(SendToSentryBaseTest):
    def assert_run(self, tty):
        self.prompt_mock.return_value = "YES"
        self.mock_isatty.return_value = tty

        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), False)
        except Exception:
            self.fail("Exception unexpectedly raised")

        if tty:
            self.raven_client_mock.assert_called()
        else:
            # Cannot prompt if not connected to TTY.
            self.raven_client_mock.assert_not_called()

        # It we have a tty, then the trace should be saved to a file and sent to sentry.
        # If we don't have a tty, then the same should happen, but the trace should
        # also be printed.
        self.error_mock.assert_called_once_with("not a SnapcraftError")
        self.exit_mock.assert_called_once_with(1)

    def test_send_tty(self):
        self.assert_run(True)
        self.assert_print_exception_called_only_tracefile(RuntimeError)

    def test_send_no_tty(self):
        self.assert_run(False)
        self.assert_print_exception_called_both_stdout_and_tempfile(RuntimeError)


class SendToSentryIsNoTest(SendToSentryBaseTest):
    def assert_run(self, tty):
        self.prompt_mock.return_value = "NO"
        self.mock_isatty.return_value = tty

        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), False)
        except Exception:
            self.fail("Exception unexpectedly raised")

        self.raven_client_mock.assert_not_called()

        # It we have a tty, then the trace should be saved to a file and sent to sentry.
        # If we don't have a tty, then the same should happen, but the trace should
        # also be printed.
        self.error_mock.assert_called_once_with("not a SnapcraftError")
        self.exit_mock.assert_called_once_with(1)

    def test_no_send_tty(self):
        self.assert_run(True)
        self.assert_print_exception_called_only_tracefile(RuntimeError)

    def test_no_send_no_tty(self):
        self.assert_run(False)
        self.assert_print_exception_called_both_stdout_and_tempfile(RuntimeError)


class SendToSentryIsAlwaysTest(SendToSentryBaseTest):
    def assert_run(self, tty):
        self.prompt_mock.return_value = "ALWAYS"
        self.mock_isatty.return_value = tty

        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), False)
        except Exception:
            self.fail("Exception unexpectedly raised")

        if tty:
            self.raven_client_mock.assert_called()

            config_path = os.path.join(
                xdg.BaseDirectory.save_config_path("snapcraft"), "cli.cfg"
            )
            self.assertThat(
                config_path,
                FileContains(
                    dedent(
                        """\
                [Sentry]
                always_send = true

                """
                    )
                ),
            )

            # It we have a tty, then the trace should be saved to a file and sent to sentry.
            self.assert_print_exception_called_only_tracefile(RuntimeError)
        else:
            # Cannot prompt if not connected to TTY.
            self.raven_client_mock.assert_not_called()

            # If we don't have a tty, the trace should be printed and saved to file.
            self.assert_print_exception_called_both_stdout_and_tempfile(RuntimeError)

        self.error_mock.assert_called_once_with("not a SnapcraftError")
        self.exit_mock.assert_called_once_with(1)

    def test_send_and_set_to_always_tty(self):
        self.assert_run(True)

    def test_send_and_set_to_always_no_tty(self):
        self.assert_run(False)


class SendToSentryAlreadyAlwaysTest(SendToSentryBaseTest):
    def test_send_as_always(self):
        config_path = os.path.join(
            xdg.BaseDirectory.save_config_path("snapcraft"), "cli.cfg"
        )
        with open(config_path, "w") as f:
            f.write(
                dedent(
                    """\
                [Sentry]
                always_send = true

                """
                )
            )

        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), True)
        except Exception:
            self.fail("Exception unexpectedly raised")

        self.assert_exception_traceback_exit_1_with_debug()
        self.raven_client_mock.assert_called()
        self.prompt_mock.assert_not_called()

    def test_send_with_config_error_does_not_save_always(self):
        self.prompt_mock.return_value = "ALWAYS"

        config_path = os.path.join(
            xdg.BaseDirectory.save_config_path("snapcraft"), "cli.cfg"
        )
        with open(config_path, "w") as f:
            f.write("bad data")

        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), True)
        except Exception:
            self.fail("Exception unexpectedly raised")

        self.assert_exception_traceback_exit_1_with_debug()
        self.raven_client_mock.assert_called()

        # Given the corruption, ensure it hasn't been written to
        self.assertThat(config_path, FileContains("bad data"))


class SendToSentryDisabledTest(SendToSentryBaseTest):
    def test_disabled_no_send(self):
        self.prompt_mock.return_value = "yes"
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_ENABLE_ERROR_REPORTING", "no")
        )

        try:
            self.call_handler(RuntimeError("not a SnapcraftError"), True)
        except Exception:
            self.fail("Exception unexpectedly raised")

        self.assert_exception_traceback_exit_1_with_debug()
        self.raven_client_mock.assert_not_called()


class _Tracefile:
    def __init__(self, test):
        self._test = test

    def __eq__(self, other):
        if not hasattr(other, "name"):
            return False

        if re.match("{}.*/trace.txt".format(tempfile.gettempdir()), other.name):
            return True

        return False


class _Stdout:
    def __init__(self, test):
        self._test = test

    def __eq__(self, other):
        if not hasattr(other, "name"):
            return False

        if other.name == "<stdout>":
            return True

        return False
